/**
 * abssqr.com Inc.
 * Copyright (c) 2017-${YEAR} All Rights Reserved.
 */
package com.abssqr.plat.biz.validate;

import com.abssqr.plat.common.facade.validation.VldNotNull;
import com.abssqr.plat.common.model.exception.AbssqrInvalidParamException;
import com.abssqr.plat.common.model.exception.AbssqrValidationException;
import com.general.system.common.util.MessageUtil;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 *
 * @author zhenxuan.luo
 * @version $Id: ValidateServiceImpl.java, v 0.1 2015年5月18日 下午7:40:35 zhenxuan.luo Exp $
 */
public class ValidateServiceImpl implements ValidateService {
    
    private final static String EMPTY_STRING = "";

    /*
     * ==================================== 内部属性
     */
    /** {@code Map<校验注解class, 校验器>} 一个校验注解对应一个校验器 */
    private final LinkedHashMap<Class<? extends Annotation>, AFieldValidator<? extends Annotation>> annoClassAValidatorMap = new LinkedHashMap<Class<? extends Annotation>, AFieldValidator<? extends Annotation>>();
    
    /** 校验注解class的set */
    private final Set<Class<? extends Annotation>> annoClassSet = new HashSet<Class<? extends Annotation>>();

    /*
     * ==================================== 类结构和注解结构的缓存
     */
    /** {@link Class#getDeclaredFields()}的cache,实测可以提升10倍的速度(内部Filed数组按照annoClassAValidatorMap的put顺序排序) */
    private final ConcurrentHashMap<String, List<Field>> fieldCache = new ConcurrentHashMap<String, List<Field>>();
    
    /** {@link Field#getAnnotations()}的cache,实测可以提升10%的速度*/
    private final ConcurrentHashMap<String, List<Annotation>> annotationCache = new ConcurrentHashMap<String, List<Annotation>>();
    
    /**
     * virtaul set method
     * 用于初始化validators
     *
     * @param fieldValidators
     */
    public void setValidators(final List<AFieldValidator<? extends Annotation>> fieldValidators) {
        //先put的Validator的优先级要高于后put的
        for (AFieldValidator<? extends Annotation> validator : fieldValidators) {
            register(validator);
        }
    }
    
    @Override
    public ValidatorError validate(final Object targetObj) {
        ValidatorError defaultResult = new ValidatorError();
        
        if (targetObj == null) {
            throw new AbssqrInvalidParamException("业务入参不能为空");
        }
        
        // 取出校验对象的所有的字段
        // 如果第一次校验，则初始化缓存
        List<Field> fields = getMarkedFields(targetObj);
        for (Field field : fields) {
            //取出每个字段的所有注解
            String key = getFieldKey(field);
            List<? extends Annotation> annotations = annotationCache.get(key);
            //这个annotations是过滤掉非校验注解,同时按照put的顺序排序了注解
            for (Annotation annotation : annotations) {
                //通过注解找到对应的校验类
                AFieldValidator<? extends Annotation> validator = this
                                                                          .getValidator(annotation
                                                                                                .annotationType());
                if (validator != null) {
                    try {
                        //校验某个字段(一个校验器一次校验一个字段)
                        FieldError fieldError = validator.validate(this, targetObj, field,
                                annotation);
                        //只要发现校验失败就直接返回
                        if (fieldError != null) {
                            defaultResult.addFieldError(fieldError);
                            break;//第一次失败后,就直接校验下一个字段
                        }
                    } catch (IllegalAccessException e) {
                        FieldError fieldError = new FieldError(targetObj.getClass().getName(),
                                                                      field.getName(), "UNKNOWN");
                        fieldError.setErrorMsg("IllegalAccessException");
                        defaultResult.addFieldError(fieldError);
                    }
                }
            }
        }
        return defaultResult;
    }
    
    /**
     * 收集所有错误的参数信息
     *
     * @param targetObject 参数对象
     * @param errors 错误
     * @return 错误参数信息
     */
    @Override
    public String collectErrorsLog(final Object targetObject, final ValidatorError errors) {
        
        if (targetObject == null || errors == null || !errors.hasErrors()) {
            return EMPTY_STRING;
        }
        
        StringBuilder sb = new StringBuilder();
        
        sb.append("").append(targetObject.getClass().getSimpleName()).append("|参数校验不通过|")
                .append("|message=[");
        
        // 添加校验不通过的所有参数名、值和不通过信息
        for (int i = 0, size = errors.getFieldErrors().size(); i < size; i++) {
            FieldError error = errors.getFieldErrors().get(i);
            sb.append(error.getFieldName()).append("=").append(error.getFieldValue()).append(":")
                    .append(error.getErrorMsg());
            if (i < size - 1) {
                sb.append(",");
            }
        }
        
        return sb.append("]").toString();
    }
    
    /**
     * 添加校验器
     * @param validator 校验器
     */
    public <A extends Annotation, AF extends AFieldValidator<A>> void register(final AF validator) {
        Class<A> annotationClass = validator.getAnnotationClass();
        AFieldValidator<?> previous = annoClassAValidatorMap.put(annotationClass, validator);
        if (previous != null) {
            throw new AbssqrValidationException("校验注解重复注册,一个校验注解注册了两个校验器" + ":" + validator);
        }
        annoClassSet.add(annotationClass);
    }
    
    /**
     * 通过注解获取校验器
     * @param annotationClass 通过注解获取校验器
     * @return 校验器
     */
    private AFieldValidator<? extends Annotation> getValidator(final Class<? extends Annotation> annotationClass) {
        return annoClassAValidatorMap.get(annotationClass);
    }
    
    /**
     * 会遍历所有父类,获取所有被校验注解标识的字段,并使用缓存
     * @param validTargetObj 需要校验的对象
     * @return 校验对象和其子类所有被校验注解标识了的字段
     */
    private List<Field> getMarkedFields(final Object validTargetObj) {
        Class<?> validTargetClass = validTargetObj.getClass();
        
        // 1 缓存中已存在，直接返回
        List<Field> cacheData = fieldCache.get(validTargetClass.getName());
        if (cacheData != null) {
            return cacheData;
        }
        
        // 2 缓存中不存在，构造缓存
        // 所有field->annotation的映射在备份的map中组装好，一次性塞入，防止因为并发而被篡改
        Map<String, List<Annotation>> bakFieldAnnotations = new HashMap<String, List<Annotation>>();
        
        //被标注的字段列表
        List<Field> markedFieldList = new ArrayList<Field>();
        
        //遍历所有父类
        Class<?> clazz = validTargetClass;
        for (; clazz != Object.class; clazz = clazz.getSuperclass()) {
            Field[] declaredFields = clazz.getDeclaredFields();//这个操作比较耗时
            for (Field field : declaredFields) {
                if (filterAndSortAnnotations(bakFieldAnnotations, field, validTargetClass)) {
                    markedFieldList.add(field);
                }
            }
        }
        
        // 3 回填缓存
        // 因为类结构在运行时是固定的，因此不考虑并发的场景
        // 先回填field->annotation的映射，再回填className->filedName的映射
        for (String key : bakFieldAnnotations.keySet()) {
            annotationCache.putIfAbsent(key, bakFieldAnnotations.get(key));
        }
        
        fieldCache.put(validTargetClass.getName(), markedFieldList);
        return markedFieldList;
    }
    
    /**
     * 过滤掉非校验注解,且将{@link VldNotNull}排序到前面,内部调用 {@link Field#getAnnotations()}方法,并使用缓存
     * 合并父类中的注解
     * @param field
     */
    private boolean filterAndSortAnnotations(final Map<String, List<Annotation>> bakFieldAnnotations,
                                             final Field field, final Class<?> paramClz) {
        
        final String key = getFieldKey(field);
        List<Annotation> annotations = bakFieldAnnotations.get(key);
        
        if (null == annotations) {
            annotations = new ArrayList<Annotation>();
        } else {
            throw new AbssqrValidationException(MessageUtil.formatMsg("类型{0}中包含父类同名字段{1}", paramClz, key));
        }
        
        boolean isMarked = fillAnnotationListFromField(annotations, field);
        
        if (isMarked) {
            bakFieldAnnotations.put(key, annotations);
        }
        
        return isMarked;
    }
    
    /**
     * 从Field中获取注解
     *
     * @param field
     * @return
     */
    private boolean fillAnnotationListFromField(final List<Annotation> annotations,
                                                final Field field) {
        Set<Class<? extends Annotation>> keySet = annoClassAValidatorMap.keySet();
        //这里是按照put的顺序来构建数组 和 过滤掉非校验注解
        boolean isMarked = false;
        for (Class<? extends Annotation> annotationClass : keySet) {
            Annotation annotation = field.getAnnotation(annotationClass);
            if (annotation != null) {
                annotations.add(annotation);
                isMarked = true;
            }
        }
        return isMarked;
    }
    
    /**
     * 组装field->annotation缓存的key
     *
     * @param field
     * @return
     */
    private String getFieldKey(final Field field) {
        return field.getDeclaringClass().getName() + "_" + field.getName();
    }
}
