/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2018 All Rights Reserved.
 */
package com.abssqr.plat.biz.shared.dispatcher.buyback;

import com.abssqr.plat.biz.shared.dispatcher.DispatcherExecutorTemp;
import com.abssqr.plat.biz.shared.handler.LoanAstBuyBackHandler;
import com.abssqr.plat.common.dal.mysql.auto.dao.TrTaskDAO;
import com.abssqr.plat.common.facade.enums.TaskStatusEnum;
import com.abssqr.plat.common.model.domain.task.TrTask;
import com.abssqr.plat.common.model.repo.tr.TrTaskRepo;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.SystemDateUtil;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;

/**
 * 资产回购交易执行器
 *
 * @author hanfei
 * @version $Id: LoanAstBuyBackExecutor.java, v 0.1 2020-02-11 5:12 PM hanfei Exp $
 */
@Component
public class LoanAstBuyBackExecutor extends DispatcherExecutorTemp<TrTask> {
    private final static Logger LOGGER       = LoggerFactory.getLogger(
            LoanAstBuyBackExecutor.class);
    // 单位（秒）
    private final static int    DEFAULT_STEP = 5;

    @Autowired
    private LoanAstBuyBackHandler loanAstBuyBackHandler;
    @Autowired
    private TrTaskRepo            trTaskRepo;

    @Override
    public TrTask doExecute(TrTask node) {
        LogUtil.info(LOGGER, "回购处理astNo[{0}]batchNo[{1}]packType[{2}]",
                node.getAstNo(), node.getBatchNo(), node.getPackType());
        boolean result = false;
        Date befDt = SystemDateUtil.getSystemDate();
        try {
            // 加锁捞取 幂等判断
            TrTask lockedNode = trTaskRepo.lockByIdNowait(node.getId());
            if (null == lockedNode || lockedNode.getTaskStatus() != TaskStatusEnum.EXECUTING) {
                LogUtil.warn(LOGGER, "资产回购任务[{0}]状态[{1}]不为执行中，可能存在并发问题", node.getId(),
                        (lockedNode != null) ? lockedNode.getTaskStatus() : "-");
                return lockedNode;
            }
            // 执行回购交易操作
            result = loanAstBuyBackHandler.astBuyBackExecute(node);
            if (result) {
                return markSuccess(node);
            } else {
                LogUtil.error(LOGGER, "资产回购任务[{0}]状态[{1}]执行失败", node.getAstNo(), node);
                return markFail(node, "资产回购执行失败");
            }
        } finally {
            Date endDt = SystemDateUtil.getSystemDate();
            LogUtil.info(LOGGER, "[LoanAstBuyBackExecutor] astId=[{0}], res={1}, elipse={2}ms",
                    node.getAstNo(), result, (endDt.getTime() - befDt.getTime()));
        }
    }

    @Autowired
    private TrTaskDAO trTaskDAO;

    /**
     * 标记成功
     *
     * @param node
     */
    @Override
    protected TrTask markSuccess(TrTask node) {
        node.setTaskStatus(TaskStatusEnum.SUCCESS);
        trTaskDAO.updateTaskStatus(node.getId(), "", TaskStatusEnum.SUCCESS.getCode());
        return node;
    }

    /**
     * 标记失败
     *
     * @param node
     * @param failMessage
     */
    @Override
    protected TrTask markFail(TrTask node, String failMessage) {
        node.setTaskStatus(TaskStatusEnum.FAILURE);
        trTaskDAO.updateTaskStatus(node.getId(), failMessage, TaskStatusEnum.FAILURE.getCode());
        return node;
    }

    /**
     * 标记异常
     *
     * @param node
     * @param failMessage
     */
    @Override
    protected TrTask markException(TrTask node, String failMessage) {
        if (StringUtils.isBlank(failMessage)) {
            failMessage = "回购交易执行失败";
        }
        node.setTaskStatus(TaskStatusEnum.INIT);
        trTaskDAO.updateTaskExep(DateTimeUtil.addSecond(SystemDateUtil.getSystemDate(), DEFAULT_STEP),
                node.getId(),
                failMessage.length() > 2000 ? failMessage.substring(0, 1990) : failMessage);
        return node;
    }

    @Override
    protected String getName() {
        return LoanAstBuyBackStarter.BUY_BACK_NAME;
    }
}
