package com.abssqr.plat.biz.shared.dispatcher.capitalAssetMatch;

import com.abssqr.plat.biz.shared.dispatcher.DispatcherExecutorTemp;
import com.abssqr.plat.biz.shared.handler.CapitalAstMatchHandler;
import com.abssqr.plat.biz.shared.handler.CapitalAstUnMatchHandler;
import com.abssqr.plat.common.facade.enums.CommonErrorCodeEnum;
import com.abssqr.plat.common.facade.enums.TaskStatusEnum;
import com.abssqr.plat.common.model.domain.task.CommonTask;
import com.abssqr.plat.common.model.exception.AbssqrBizException;
import com.abssqr.plat.common.model.repo.task.CommonTaskRepository;
import com.abssqr.plat.core.service.accounting.OrgAcctDayComponent;
import com.general.system.common.model.exception.BaseCommonException;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.MessageUtil;
import com.general.system.common.util.SystemDateUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.transaction.support.TransactionTemplate;

import java.util.Date;

/**
 * 资金资产绑定/解绑
 */
@Component
public class PlanCapitalAssertMatchExecutor extends DispatcherExecutorTemp<CommonTask> {
    private final static Logger LOGGER       = LoggerFactory.getLogger(
            PlanCapitalAssertMatchExecutor.class);
    @Autowired
    CommonTaskRepository commonTaskRepository;
    @Autowired
    CapitalAstMatchHandler capitalAstMatchHandler;
    @Autowired
    CapitalAstUnMatchHandler capitalAstUnMatchHandler;
    @Autowired
    private OrgAcctDayComponent orgAcctDayComponent;
    @Autowired
    TransactionTemplate transactionTemplate;
    
    @Override
    protected CommonTask doExecute(CommonTask node) {
        Date absAccountDate = orgAcctDayComponent.getTrfAcctDay().getStandardDate();
        StringBuilder failMessage=new StringBuilder();
        boolean isSucc=true;
        switch (node.getTaskType()) {
                //解绑任务
                case CAPTITALASSERT_UNMATCH: {
                    LogUtil.info(LOGGER, "执行资金资产解绑-start", node.getContext());
                    String allocNo = node.getBizNo();
                    String planNo = node.getContext();
                    try{
                        isSucc=transactionTemplate.execute(status -> {
                            return capitalAstUnMatchHandler.execute(planNo,allocNo,absAccountDate);
                        });
                    }catch (Exception e){
                        LogUtil.error(LOGGER, "计划编号[{0}],兑付单编号[{1}]执行资金资产解绑失败错误信息:[3]",planNo,allocNo,e.getMessage());
                        isSucc=false;
                        failMessage.append(e.getMessage());
                    }
                    markStatus(isSucc,node,failMessage.toString());
                    LogUtil.info(LOGGER, "执行资金资产解绑-end", node.getContext());
                    break;
                }
                //解绑匹配任务
                case UNMATCH_CAPTITALASSERT_MATCH: {
                    try{
                        isSucc=transactionTemplate.execute(status -> {
                            return doMatch(node);
                        });
                    }catch (Exception e){
                        LogUtil.error(LOGGER, "计划编号[{0}]执行资金资产匹配败错误信息:[3]",node.getContext(),e.getMessage());
                        isSucc=false;
                        failMessage.append(e.getMessage());
                    }
                    markStatus(isSucc,node,failMessage.toString());
                    break;
                }
                //同步匹配任务
                case SYNC_CAPTITALASSERT_MATCH: {
                    try{
                        isSucc=transactionTemplate.execute(status -> {
                            return doMatch(node);
                        });
                    }catch (Exception e){
                        LogUtil.error(LOGGER, "计划编号[{0}]执行资金资产匹配败错误信息:[3]",node.getContext(),e.getMessage());
                        isSucc=false;
                        failMessage.append(e.getMessage());
                    }
                    markStatus(isSucc,node,failMessage.toString());
                    break;
                }
                default:{
                    throw new BaseCommonException("不支持的任务类型:"+node.getTaskType().getCode());
                }
            }
       return node;
    }
    private boolean doMatch(CommonTask node){
        LogUtil.info(LOGGER, "执行解绑匹配-start", node.getContext());
        String bizNo = node.getBizNo();
        String[] split = bizNo.split("-");
        if(split==null){
            throw new AbssqrBizException(
                    MessageUtil.formatMsg("Id为[{0}]的CommonTaskBizNo有误 ", node.getId()),
                    CommonErrorCodeEnum.INVALID_PARAM, false);
        }
        String planNo=node.getContext();
        String accDateStr=split[2];
        Date acctDate = DateTimeUtil.parseFromYMD(accDateStr);

        boolean ret= capitalAstMatchHandler.execute(planNo,acctDate);
        LogUtil.info(LOGGER, "执行解绑匹配-end", node.getContext());
        return ret;
    }
    private void markStatus(Boolean status,CommonTask node,String errorMessage){
        if(status){
            markSuccess(node);
            return;
        }
        markFail(node,errorMessage);
    }
    @Override
    protected CommonTask markSuccess(CommonTask node) {
        //为保证数据一直性,讲成功设置
        commonTaskRepository.updateTaskStatus(node.getId(), TaskStatusEnum.SUCCESS, TaskStatusEnum.EXECUTING, "");
        return node;
    }

    @Override
    protected CommonTask markFail(CommonTask node, String failMessage) {
        Date nextTime = DateTimeUtil.addSecond(SystemDateUtil.getSystemDate(), 60);
        String msg = failMessage.length() > 2000 ? failMessage.substring(0, 2000) : failMessage;
        commonTaskRepository.updateTaskStatus(node.getId(), TaskStatusEnum.IGNORE, TaskStatusEnum.EXECUTING, "");
        commonTaskRepository.updateTaskExep(nextTime, node.getId(), msg);
        return node;
    }

    @Override
    protected CommonTask markException(CommonTask node, String failMessage) {
        return markFail(node, failMessage);
    }

    @Override
    protected String getName() {
        return "PlanCapitalAssertMatchExecutor";
    }
}
