/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2020 All Rights Reserved.
 */

package com.abssqr.plat.biz.shared.handler;

import com.abssqr.plat.biz.shared.scheduler.job.capitalAssertMatch.CapitalAssertMatchJob;
import com.abssqr.plat.common.dal.mysql.auto.paging.CapitalAstMatchPage;
import com.abssqr.plat.common.dal.mysql.auto.resultmap.AstScale;
import com.abssqr.plat.common.facade.constants.LoanExtInfConstants;
import com.abssqr.plat.common.facade.enums.EventCodeEnum;
import com.abssqr.plat.common.facade.enums.OwnStatusEnum;
import com.abssqr.plat.common.facade.enums.TaskStatusEnum;
import com.abssqr.plat.common.facade.enums.TransCodeEnum;
import com.abssqr.plat.common.facade.enums.YNEnum;
import com.abssqr.plat.common.facade.model.plan.InvestStructure;
import com.abssqr.plat.common.facade.model.plan.PlanAcctTitle;
import com.abssqr.plat.common.model.domain.ast.Loan;
import com.abssqr.plat.common.model.domain.ast.LoanAstGroup;
import com.abssqr.plat.common.model.domain.ast.LoanBak;
import com.abssqr.plat.common.model.domain.job.JobControl;
import com.abssqr.plat.common.model.domain.loanTrans.LoanTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.TransInfoTemplate;
import com.abssqr.plat.common.model.repo.ast.LoanBakRepo;
import com.abssqr.plat.common.model.repo.ast.LoanRepository;
import com.abssqr.plat.common.model.repo.job.JobControlRepository;
import com.abssqr.plat.common.model.repo.loanTrans.LoanLogRepository;
import com.abssqr.plat.common.model.repo.plan.PlanAcctTitleRepo;
import com.abssqr.plat.common.model.repo.plan.PlanAllocRepo;
import com.abssqr.plat.common.model.repo.tr.TrTaskRepo;
import com.abssqr.plat.common.model.seq.SequenceUtil;
import com.abssqr.plat.common.util.profiler.PerformTrace;
import com.abssqr.plat.common.util.profiler.PerformTraceContext;
import com.abssqr.plat.core.service.accounting.OrgAcctDayComponent;
import com.abssqr.plat.core.service.ast.LoanAstAssembler;
import com.abssqr.plat.core.service.plan.PlanInvestStructureManager;
import com.abssqr.plat.core.service.system.TimeStampUtil;
import com.general.system.common.model.Money;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.SystemDateUtil;
import com.general.system.common.util.VarChecker;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.transaction.support.TransactionTemplate;

import java.math.BigDecimal;
import java.util.Comparator;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import static com.abssqr.plat.common.model.enums.IdTypeEnum.CAP_AST_MATCH;

/**
 * 资金资产匹配
 *
 * @author xiachenxiang
 * @version com.abssqr.plat.biz.shared.handler: CapitalAstMatchHandler.java, v 0.1 2020-02-03 2:22 PM xiachenxiang Exp $
 */
@Component
public class CapitalAstMatchHandler {
    private final Logger LOGGER = LoggerFactory.getLogger(getClass());

    @Autowired
    private TrTaskRepo                 trTaskRepo;
    @Autowired
    private PlanInvestStructureManager planInvestStructureManager;
    @Autowired
    private PlanAcctTitleRepo          planAcctTitleRepo;
    @Autowired
    private LoanBakRepo                loanBakRepo;
    @Autowired
    private LoanRepository             loanRepository;
    @Autowired
    private TransactionTemplate        transactionTemplate;
    @Autowired
    private LoanAstAssembler           loanAstAssembler;
    @Autowired
    private LoanLogRepository          loanLogRepository;
    @Autowired
    private PlanAllocRepo              planAllocRepo;
    @Autowired
    private JobControlRepository       jobControlRepository;
    @Autowired
    private CapitalAssertMatchJob      capitalAssertMatchJob;
    @Autowired
    private OrgAcctDayComponent orgAcctDayComponent;


    private int queryLimit = 2000;

    /**
     * @param planNo        计划编号
     * @param taskCreateDay 任务创建日
     * @return
     */
    @PerformTrace(name = "CapitalAstMatchHandler")
    public boolean execute(String planNo, Date taskCreateDay) {
        // 1. 判断是否可进行资金匹配
        // 判断转让、回购任务是否全部完成 ignore、init、exec
        Long cnt = trTaskRepo.cntTrTaskByTfrOut(planNo, Lists.newArrayList(TaskStatusEnum.IGNORE.getCode(), TaskStatusEnum.INIT.getCode(), TaskStatusEnum.EXECUTING.getCode()), null);
        VarChecker.checkArgument(cnt.equals(0L), "[CapitalAstMatchHandler] - [{0}]转让任务未完成", planNo);
        // 判断会计日是否相等
        Date standardDate = orgAcctDayComponent.getTrfAcctDay().getStandardDate();
        if (!DateTimeUtil.isSameDay(standardDate, taskCreateDay)) {
            LogUtil.error(LOGGER, "[CapitalAstMatchHandler] - [{0}]资金匹配计划匹配日期为[{1}],当前会计日为[{2}],跳过", planNo, DateTimeUtil.formatYMD(taskCreateDay), DateTimeUtil.formatYMD(standardDate));
            return true;
        }

//        LogUtil.info(LOGGER, "[CapitalAstMatchHandler] - [{0}]开始资金资产匹配", planNo);

        // 2. 计算每个投资层级资金应该匹配多少资产
        // 可用资金=投资层级规模-待分配本金金额（已审批）
        // 获取投资结构信息
        List<InvestStructure> investStructureList = planInvestStructureManager.selectByPlanNo(planNo);
        // 升序
        investStructureList.sort(Comparator.comparing(InvestStructure::getLayerEndDate));
        Map<String, Money> needMatchBalMap = this.calcNeedMatchMap(planNo, investStructureList, taskCreateDay);
        if (MapUtils.isEmpty(needMatchBalMap)) {
            LogUtil.info(LOGGER, "[CapitalAstMatchHandler] - [{0}]计划剩余需匹配投资结构为0", planNo);
            return true;
        }

        // 3. 执行资金资产匹配
        boolean match = executeMatch(planNo, needMatchBalMap, investStructureList, taskCreateDay);

        LogUtil.info(LOGGER, "[CapitalAstMatchHandler] - [{0}]结束资金资产匹配 - result [{1}]", planNo, match);
        return match;
    }

    /**
     * 执行资金匹配
     *
     * @param planNo              计划编号
     * @param needMatchBalMap     应匹金额
     * @param investStructureList 投资结构
     * @param taskCreateDay       任务创建日期
     * @return
     */
    private boolean executeMatch(String planNo, Map<String, Money> needMatchBalMap, List<InvestStructure> investStructureList, Date taskCreateDay) {

        try {
            PerformTraceContext.enter("CapitalAstMatchHandler.executeMatch");
            // 资金资产匹配
            CapitalAstMatchPage capitalAstMatchPage = new CapitalAstMatchPage();
            capitalAstMatchPage.setCurrPageNo(1);
            capitalAstMatchPage.setLimit(queryLimit);
            capitalAstMatchPage.setLenderOrgCode(planNo);
            boolean match = true;
            do {
                match = this.capitalAstMatch(capitalAstMatchPage, needMatchBalMap, investStructureList, taskCreateDay, planNo);
                if (!match) {
                    break;
                }
            } while (capitalAstMatchPage.hasNextPage());
            // 判断是否存在未匹配的资产
            capitalAstMatchPage = loanRepository.queryForCapitalAstMatch(capitalAstMatchPage);
            if (capitalAstMatchPage.getTotal() > 0) {
                LogUtil.warn(LOGGER, "[CapitalAstMatchHandler] - [{0}]结束资金资产匹配失败,剩余应批资产大于0", planNo);
                match = false;
            }
            return match;
        } finally {
            PerformTraceContext.leave();
        }
    }

    /**
     * 计算每个投资层级资金应该匹配多少资产
     *
     * @param planNo              计划编号
     * @param investStructureList 投资结构
     * @param taskCreateDay       任务创建日期
     * @return
     */
    private Map<String, Money> calcNeedMatchMap(String planNo, List<InvestStructure> investStructureList, Date taskCreateDay) {
        // 获取本金科目
        List<PlanAcctTitle> planAcctTitleList = planAcctTitleRepo.getByEntityNo(planNo);
        //  获取已审批未兑付的金额
        Map<String, Money> willPayMoneyMap = planAllocRepo.queryApprovedNotPayMoneyGroupByLayerNo(planNo);

        Map<String, Money> balMap = Maps.newHashMap();
        Money investStructureBalSum = new Money();
        planAcctTitleList.forEach(item -> {
            if (item.getTitleType().isInvest()) {
                Money money = item.getBal().clone();
                Money willPayMoney = willPayMoneyMap.getOrDefault(item.getRefNo(), new Money());
                money.subtractFrom(willPayMoney);
                // 负保护
                VarChecker.checkArgument(!money.isNegtive(), "[CapitalAstMatchHandler] - [{0}]投资结构剩余应匹规模小于0", item.getRefNo());
                balMap.put(item.getRefNo(), money);
                investStructureBalSum.addTo(money);
            }
        });
        VarChecker.checkArgument(investStructureList.size() == balMap.size(), "[CapitalAstMatchHandler] - [{0}]投资结构与本金科目不一致", planNo);
        // 2.2 计算资金比例
        // 计算资金比例
        Map<String, BigDecimal> rateMap = this.calcCapistalRate(investStructureList, balMap, investStructureBalSum, planNo);
        Money astScaleSum = new Money();
        // 获取信托资产规模
        Map<String, Money> astScaleMap = this.getAstScaleMap(taskCreateDay, planNo, astScaleSum);

        // 计算目标匹配余额
        this.calcNeedMatchBal(investStructureList, astScaleSum, astScaleMap, rateMap, investStructureBalSum, planNo, balMap, true);
        // 计算资金比例
        rateMap = this.calcCapistalRate(investStructureList, balMap, investStructureBalSum, planNo);
        // 计算目标匹配余额
        Map<String, Money> needMatchBalMap = this.calcNeedMatchBal(investStructureList, astScaleSum, astScaleMap, rateMap, investStructureBalSum, planNo, balMap, false);
        return needMatchBalMap;
    }

    /**
     * @param taskCreateDay
     * @param planNo
     * @return
     */
    private Map<String, Money> getAstScaleMap(Date taskCreateDay, String planNo, Money astScaleSum) {

        List<AstScale> astScaleList = loanBakRepo.getAstScale(taskCreateDay, planNo);
        Map<String, Money> astScaleMap = Maps.newHashMap();
        astScaleList.forEach(item -> {
            astScaleSum.addTo(item.getPrinBal());
            astScaleMap.put(item.getCapitalNo(), item.getPrinBal());
        });
        return astScaleMap;
    }

    /**
     * 资金资产匹配
     *
     * @param capitalAstMatchPage 分页
     * @param needMatchBalMap     匹配余额
     * @param investStructureList 投资结构信息
     * @param execDate            任务生成的会计日
     * @return 匹配结果
     */
    private boolean capitalAstMatch(CapitalAstMatchPage capitalAstMatchPage, Map<String, Money> needMatchBalMap,
                                    List<InvestStructure> investStructureList, Date execDate, String planNo) {
        boolean match = true;
        //增加Job控制
        JobControl jobControl = jobControlRepository.selectByJobName(capitalAssertMatchJob.getJobName());
        if (jobControl.getIsEff() == YNEnum.N) {
            LogUtil.error(LOGGER, "计划[{0}]匹配失败，资金资产匹配任务运行中被关闭", planNo);
            return false;
        }
        capitalAstMatchPage = loanRepository.queryForCapitalAstMatch(capitalAstMatchPage);
        List<String> astNoList = capitalAstMatchPage.getDatas();
        for (String astNo : astNoList) {

            try {
                PerformTraceContext.enter("CapitalAstMatchHandler.capitalAstmatch");
                match = transactionTemplate.execute(status -> this.matchCapital(astNo, investStructureList, needMatchBalMap, execDate, planNo));
            } finally {
                PerformTraceContext.leave();
            }

            if (!match) {
                break;
            }

        }
        return match;
    }

    /**
     * 匹配资金
     *
     * @param astNo               资产编号
     * @param investStructureList 投资结构 按到期日升序
     * @param needMatchBalMap     投资结构对应剩余应匹配金额
     * @param execDate            任务生成日期
     * @return
     */
    private boolean matchCapital(String astNo, List<InvestStructure> investStructureList,
                                 Map<String, Money> needMatchBalMap, Date execDate, String planNo) {
        // 机构会计日
//        Date trfDate = orgAcctDayComponent.getTrfAcctDay().getStandardDate();
        
        Loan loan = loanAstAssembler.lockByAstNo(astNo, execDate);
        // 判断日期
        if (DateTimeUtil.calcDayDiff(loan.getLastUpdateDate(), execDate) > 0) {
            LogUtil.error(LOGGER, "[{0}]资金匹配失败,[{1}]资产更新日期[{2}]大于任务生成日期[{3}]", planNo, loan.getAstNo(), loan.getLastUpdateDate(), execDate);
            return false;
        }
        // 判断是否全部转让
        if (loan.getAstOrg().getAstOwnership().getOwnStatus() == OwnStatusEnum.TRANS_OUT) {
            LogUtil.info(LOGGER, "[{0}]资金匹配跳过,[{1}]资产已全部转让,可能存在转让任务,跳过该资产", planNo, loan.getLastUpdateDate(), execDate);
            return true;
        }
        for (InvestStructure investStructure : investStructureList) {
            // 判断投资结构是否需要匹配
            if (needMatchBalMap.get(investStructure.getLayerNo()) == null) {
                continue;
            }
            Money needMatchBal = needMatchBalMap.get(investStructure.getLayerNo());
            if (!needMatchBal.isGreaterThanZero()) {
                continue;
            }
            Map<String, String> extInf = loan.getExtInf();
            String bfCapitalNo = extInf.get(LoanExtInfConstants.BF_CAPITAL_NO);
            if (StringUtils.isNotEmpty(bfCapitalNo) && StringUtils.equals(bfCapitalNo, investStructure.getLayerNo())) {
                continue;
            }
            // 资产到期日 小于等于 投资结构到期日
            if (DateTimeUtil.calcDayDiff(investStructure.getLayerEndDate(), loan.getAstPeriod().getEndDate()) >= 0) {
                // 匹配成功
                LoanAstGroup loanAstGroup = loan.filterLoanAstByOrg(loan.getAstOrg().getAstOwnership().getLenderOrgCode(), loan.getTermCnt());
                Money loanAstBal = new Money();
                loanAstGroup.getLoanAsts().forEach(item -> {
                    loanAstBal.addTo(item.getPrinBal());
                });

                // 更新目标匹配余额
                needMatchBal.subtractFrom(loanAstBal);

                String batchNo = SequenceUtil.genId(CAP_AST_MATCH);
                String timeStamp = TimeStampUtil.getDefaultTimeStamp();
                TransInfoTemplate transInTemplate = new TransInfoTemplate();
                transInTemplate.setTransTime(SystemDateUtil.getSystemDate());
                transInTemplate.setAccountDate(execDate);
                transInTemplate.setEventCode(EventCodeEnum.CAP_AST_MATCH);
                transInTemplate.setOutBizNo(batchNo);
                transInTemplate.setBizNo(SequenceUtil.genSeqNo(TransCodeEnum.CAP_AST_MATCH));
                transInTemplate.setTimestamp(timeStamp);
                // 构造交易详情
                LoanTransDetail loanTransDetail = LoanTransDetail.createLoanAstTxnDtl(loan, transInTemplate);
                // 更新资金编号
                loanTransDetail.setCapitalInfo(investStructure.getLayerNo(), investStructure.getLayerName());
                loanTransDetail.processBeforeUpdate();
                // 更新信息
                loanAstAssembler.store(loan, execDate, false);
                List<LoanBak> loanBaks = LoanTransDetail.buildBaks(loan, transInTemplate.getAccountDate());
                loanRepository.backup(loanBaks);
                return true;
            }

        }
        return false;
    }


    /**
     * 计算匹配余额
     *
     * @param investStructureList   投资结构列表
     * @param astScaleSum           资产总规模
     * @param astScaleMap           资产规模
     * @param rateMap               资金比例
     * @param investStructureBalSum 投资结构总余额
     * @param planNo                计划编号
     * @param balMap                投资结构资金余额
     * @param removeIfNegtive       匹配余额小于等于0时是否移除 false 不移除 true 移除
     * @return
     */
    private Map<String, Money> calcNeedMatchBal(List<InvestStructure> investStructureList, Money astScaleSum, Map<String, Money> astScaleMap, Map<String, BigDecimal> rateMap, Money investStructureBalSum, String planNo, Map<String, Money> balMap, boolean removeIfNegtive) {
        Map<String, Money> needMatchBalMap = Maps.newHashMap();
        Iterator<InvestStructure> iterator = investStructureList.iterator();
        while (iterator.hasNext()) {
            InvestStructure investStructure = iterator.next();
            if (balMap.get(investStructure.getLayerNo()) == null) {
                continue;
            }
            Money needMatchBal = astScaleSum.multiply(rateMap.get(investStructure.getLayerNo()));
            Money alreadyMatchMoney = astScaleMap.getOrDefault(investStructure.getLayerNo(), new Money());

            // 判断目标匹配余额，这边会修改needMatchBal的值
            needMatchBal.subtractFrom(alreadyMatchMoney);
            if (!needMatchBal.isGreaterThanZero() && removeIfNegtive) {
                investStructureBalSum.subtractFrom(balMap.get(investStructure.getLayerNo()));
                balMap.remove(investStructure.getLayerNo());
                continue;
            }
            needMatchBalMap.put(investStructure.getLayerNo(), needMatchBal);
        }
        return needMatchBalMap;
    }

    /**
     * 计算资金比例
     *
     * @param investStructureList 投资结构
     * @param balMap              余额
     * @param sum                 总额
     * @return
     */
    private Map<String, BigDecimal> calcCapistalRate(List<InvestStructure> investStructureList, Map<String, Money> balMap, Money sum, String planNo) {
        // 非0保护
        VarChecker.checkArgument(sum.isGreaterThanZero(), "[CapitalAstMatchHandler] - [{0}]投资结构总剩余规模为0", planNo);
        Map<String, BigDecimal> rateMap = Maps.newHashMap();
        investStructureList.forEach(item -> {
            Money bal = balMap.getOrDefault(item.getLayerNo(), new Money(0));
            BigDecimal rate = bal.getAmount().divide(sum.getAmount(), 2, BigDecimal.ROUND_HALF_UP);
            rateMap.put(item.getLayerNo(), rate);
        });
        return rateMap;
    }

}
