/**
 * abssqr.com Inc. 
 * Copyright (c) 2017-2019 All Rights Reserved. 
 */
package com.abssqr.plat.biz.shared.handler;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.date.DateUtil;
import com.abssqr.plat.common.facade.enums.AbssqrNoticeTypeEnum;
import com.abssqr.plat.common.facade.enums.PlanAllocStatusEnum;
import com.abssqr.plat.common.facade.enums.PlanTypeEnum;
import com.abssqr.plat.common.facade.model.plan.PlanAcctTitle;
import com.abssqr.plat.common.facade.model.plan.PlanAlloc;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.facade.model.plan.PlanTransferRule;
import com.abssqr.plat.common.model.domain.plan.PlanAstLog;
import com.abssqr.plat.common.model.repo.plan.PlanAcctTitleRepo;
import com.abssqr.plat.common.model.repo.plan.PlanAllocRepo;
import com.abssqr.plat.common.model.repo.plan.PlanTransferRuleRepo;
import com.abssqr.plat.core.service.notice.AbssqrNoticeManager;
import com.abssqr.plat.core.service.plan.PlanAllocAmtCalcManager;
import com.abssqr.plat.core.service.plan.PlanAstManager;
import com.general.system.common.model.Money;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.MessageUtil;
import com.general.system.tool.util.ToolUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.stream.Collectors;

/**
 * 计算计划每日为分配兑付 预留冻结金额
 * @author zhenxuan.luo
 * @version com.abssqr.plat.biz.shared.handler: PlanCashFrozenHandler.java, v 0.1 2019-06-13 15:39 zhenxuan.luo Exp $ 
 */
@Component
public class PlanCashFrozenHandler {
    
    private final static Logger LOGGER = LoggerFactory.getLogger(PlanCashFrozenHandler.class);
    
    private final static int                  DAY_OFFSET = 15;
    @Autowired
    private AbssqrNoticeManager     abssqrNoticeManager;
    @Autowired
    private PlanAstManager          planAstManager;
    @Autowired
    private PlanAllocRepo           planAllocRepo;
    @Autowired
    private PlanTransferRuleRepo    planTransferRuleRepo;
    @Autowired
    private PlanAllocAmtCalcManager planAllocAmtCalcManager;
    @Autowired
    private PlanAcctTitleRepo       planAcctTitleRepo;
    
    /**
     * 资金冻结：需要在外部起事务，保证冻结和任务状态一致，避免重复冻结
     * @param planEntity 计划
     * @param curDate  转让会计日
     */
    public void execute(PlanBase planEntity, Date curDate) {
        
        LogUtil.info(LOGGER, "执行{0}[{1}]水位冻结任务", planEntity.getPlanType().getDesc(), planEntity.getPlanNo());
        
        // 1. 查找未来最近的分配计划，考虑最近期限（15天）之内多个分配计划日期较近的全部进行判断处理
        List<PlanAlloc> planAllocList = planAllocRepo.queryUnassigneAllocByPlanNo(planEntity.getPlanNo());
        if (ToolUtil.isEmpty(planAllocList)) {
            LogUtil.warn(LOGGER, "没有找到{0}[{1}]的最近未来分配计划, 当前时间[{1}]",
                    planEntity.getPlanType().getDesc(), planEntity.getPlanNo(), curDate);
            return;
        }
        
        // 2. 判断分配当期是否处于循环期，若不属于循环期，不做资金预留
        if (!this.filteNeedFrozenAllocs(curDate, planAllocList, planEntity)) {
            return;
        }
        
        // 3. 计算分配日需要分配的资金 填充需要分配的本金 费用 等金额
        List<PlanAcctTitle> planAcctTitles = planAcctTitleRepo.getByEntityNo(planEntity.getPlanNo());
        planAllocAmtCalcManager.calcTriAmtByScheduler(planEntity, planAllocList, planAcctTitles, true);
        
        // 计算合计总共需要兑付金额
        Money totalAllocAmt = new Money();
        // 多个分配兑付日
        String clearDates = CollectionUtil.join(planAllocList.stream().map(alloc -> {
            totalAllocAmt.addTo(alloc.calcTotalTrialAmt());
            return DateUtil.formatDate(alloc.getClearDate());
        }).collect(Collectors.toList()), ",");
        
        // 计算并执行每日冻结水位资金
        PlanAstLog planAstLog = planAstManager.calcFrozen(curDate, planEntity.getPlanNo(), planAllocList, clearDates);
        
        // 4. 计算近30天的日均还款
        Date startDate = DateTimeUtil.addDays(curDate, -31);
        Date endDate = DateTimeUtil.addDays(curDate, -1);
        startDate = (startDate.compareTo(planEntity.getEffectDate()) > 0) ? startDate : planEntity.getEffectDate();
    
        /**
         * 获取每日变更水位的还款金额
         * 放款池计算还款包含了承接池的还款-- 因为承接池还会转让将资金转回放款池，所以暂时不计
         */
        Money avgRepoAmt = planAstManager.getRepayTransAmtStat(
                planEntity.getPlanNo(), startDate, endDate, true);
        
        if (planAstLog.getFrozenCashAmt().greaterThan(avgRepoAmt)
            && planAstLog.getFrozenCash().greaterThan(planAstLog.getCash())) {
            // 提醒优化，可能计划本来就有很多钱，这个时候还款金额是不是足够其实不重要
            String msg = MessageUtil.formatMsg(
                    "分配计划日期为[{0}]，存在兑付风险，待兑付金额{1}元。平均每日还款金额{2}元，今日冻结金额{3}元，计划资金余额{4}元，计划所有已冻结金额{5}元"
                    , clearDates, totalAllocAmt, avgRepoAmt, planAstLog.getFrozenCashAmt(),
                    planAstLog.getCash(), planAstLog.getFrozenCash());
            // 插入一条通知，每日还款金额小于每日冻结金额
            abssqrNoticeManager.sendBusinessNotice(planEntity, planEntity.getPlanNo(),
                    AbssqrNoticeTypeEnum.PlanAllocRisk,  msg, null);
            
            LogUtil.warn(LOGGER, msg);
        }
    }
    
    /**
     * 判断是否分配计划是否需要冻结资金
     * @param curDate
     * @param planAllocs
     * @param planBase
     * @return
     */
    private boolean filteNeedFrozenAllocs(Date curDate, List<PlanAlloc> planAllocs, PlanBase planBase) {
        String planNo = planBase.getPlanNo();
        Date stopTrDate = null;
        // 承接池进行判断循环结构及期限
        if (planBase.getPlanType() == PlanTypeEnum.PLAN){
            // 查询计划转让规则
            PlanTransferRule transferRules = planTransferRuleRepo.getByPlanNo(planNo);
            if (null == transferRules) {
                LogUtil.warn(LOGGER, "没有找到计划[{0}]的转让规则", planNo);
                return false;
            }
            // 非循环结构
            if (!transferRules.getCycle()) {
                LogUtil.info(LOGGER, "计划[{0}]不需要冻结预留资金水位，计划非循环结构[{1}]", planNo, transferRules.getCycle());
                return false;
            }
            
            // 循环结果，判断循环期
            if (DateTimeUtil.calcDayDiff(curDate, transferRules.getStopTrDate()) >= 0) {
                // 当前日期已经停止循环
                LogUtil.info(LOGGER, "计划[{0}]不需要冻结预留资金水位，当期日期[{1}]大于等于循环期[{2}]", planNo, curDate,
                        transferRules.getStopTrDate());
                return false;
            }
            // 转让终止日
            stopTrDate = transferRules.getStopTrDate();
        }
        Iterator<PlanAlloc> it = planAllocs.iterator();
        while (it.hasNext()) {
            PlanAlloc planAlloc = it.next();
            if (planAlloc.getStatus() == PlanAllocStatusEnum.ASSIGNED) {
                // 剔除已分配
                it.remove();
                continue;
            }
            if (DateTimeUtil.calcDayDiff(planAlloc.getClearDate(), curDate) > DAY_OFFSET) {
                // 距离结算日提前15天预留资金
                LogUtil.info(LOGGER, "计划[{0}]不需要冻结预留资金水位，当期日期[{1}]距离最近分配计划结算日[{2}]大于[{3}]天",
                        planNo, curDate,
                        planAlloc.getClearDate(),
                        DAY_OFFSET);
                //移除该对象
                it.remove();
                continue;
            }
            
            if (stopTrDate != null && DateTimeUtil.calcDayDiff(stopTrDate, planAlloc.getClearDate()) < 0) {
                // 分配计划结算日在循环期内，需要冻结
                LogUtil.info(LOGGER, "计划[{0}]需要冻结预留资金水位，分配计划[{1}]结算日[{2}]小于循环期[{3}]", planNo,
                        planAlloc.getAllocNo(),
                        planAlloc.getClearDate(),
                        stopTrDate);
                //移除该对象
                it.remove();
                continue;
            }
        }
        
        if (ToolUtil.isNotEmpty(planAllocs)) {
            // 剩余场景：停止循环日期大于当前日期 且 停止循环日期小于分配计划结算日。保险期间，做资金预留
            return true;
        } else {
            return false;
        }
    }
    
}