/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2019 All Rights Reserved.
 */

package com.abssqr.plat.biz.shared.scheduler.job.keeper.node;

import com.abssqr.plat.biz.shared.scheduler.job.sync.ast.ZLOriginalAstSyncJob;
import com.abssqr.plat.common.facade.enums.AbssqrNoticeTypeEnum;
import com.abssqr.plat.common.facade.enums.PlanTaskTypeEnum;
import com.abssqr.plat.common.facade.enums.SyncTaskTypeEnum;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.model.domain.plan.PlanTaskLog;
import com.abssqr.plat.common.model.domain.task.ctr.AstSyncTaskCtr;
import com.abssqr.plat.common.model.domain.task.ctr.AstSyncTaskCtrLog;
import com.abssqr.plat.common.model.repo.plan.PlanTaskLogRepo;
import com.abssqr.plat.common.model.repo.task.AstSyncTaskCtrRepo;
import com.abssqr.plat.common.model.seq.SequenceUtil;
import com.abssqr.plat.common.model.wk.repo.WkDhccBatchRepo;
import com.abssqr.plat.common.util.base.Day;
import com.general.system.common.model.IEnum;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.MessageUtil;
import com.general.system.tool.util.ToolUtil;
import org.apache.commons.collections.CollectionUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.transaction.support.TransactionTemplate;

import java.util.Date;
import java.util.List;
import java.util.Set;

/**
 * 资产同步 异常监控
 * <p>
 * 到deadline为止都不存在成功的同步任务
 *
 * @author xiachenxiang
 * @version com.abssqr.plat.biz.shared.scheduler.job.keeper.node: AstSync.java, v 0.1 2019-09-18 11:00 AM xiachenxiang Exp $
 */
@Component
public class AstSyncJobKeeper extends AbstractKeeperNode {

    @Autowired
    private AstSyncTaskCtrRepo astSyncTaskCtrRepo;

    @Autowired
    private WkDhccBatchRepo wkDhccBatchRepo;

    @Autowired
    private PlanTaskLogRepo planTaskLogRepo;

    @Autowired
    private TransactionTemplate transactionTemplate;

    @Value("${wkabs.astsync.zhaolian.deadline:18 0 0}")
    private String deadline;


    @Override
    protected Void doHandle(Date currentTime) {
        LogUtil.info(LOGGER, "五矿资产同步监控-start");
        currentTime = getCurrentStanderDate();
        List<AstSyncTaskCtr> astSyncTaskCtrs = astSyncTaskCtrRepo.getAstSyncTaskCtrs();
        // 判断是否大于指定时间
        Date deadlineTime = getDeadLineTime(currentTime, deadline);
        if (ToolUtil.isNotEmpty(astSyncTaskCtrs) && currentTime.compareTo(deadlineTime) > 0) {
            for (AstSyncTaskCtr astSyncTaskCtr : astSyncTaskCtrs){
                // 获取五矿资产数据准备标记
                boolean isDataReady = wkDhccBatchRepo.checkDataReady(
                        astSyncTaskCtr.getExtData().getCtrFlg(), new Day(currentTime).getYesterday());
                if (!isDataReady) {
                    String msg = MessageUtil.formatMsg("到[{0}]为止,机构[{1}]同步数据仍未准备完成，请排查具体原因!",
                            DateTimeUtil.formatYMDHMSSLong(currentTime), astSyncTaskCtr.getOrgCode());
                    // 获取用户
                    Set<String> receivers = getAdminReceivers();
                    sendSystemNotice("五矿资产同步异常告警", astSyncTaskCtr.getOrgCode(), msg, receivers);
                }
            }
        }
        
        Date currStandDate = getCurrentStanderDate();
        Date acctDate= orgAcctDayComponent.getTrfAcctDay().getStandardDate();
        // 判断时间上午9点 校验上一日 AST_SYNC、AST_MOD任务完成
        final String yesterdaySyncAndModDeadLine="9 0 0";
        if(currentTime.compareTo(getDeadLineTime(currentTime, yesterdaySyncAndModDeadLine))>0) {
            transactionTemplate.execute(status -> {
                if (checkCanNotice(currStandDate, PlanTaskTypeEnum.LAS_SYN_MOD_CHECK)) {
                    buildAndSendMessage(astSyncTaskCtrs, DateTimeUtil.addDays(acctDate, -1));
                }
                return true;
            });
        }
        final String todaySyncAndModDeadLine="22 0 0";
        // 判断时间晚10点 校验当前系统日 AST_SYNC、AST_MOD任务完成
        if(currentTime.compareTo(getDeadLineTime(currentTime, todaySyncAndModDeadLine))>0){
            transactionTemplate.execute(status -> {
                if (checkCanNotice(currStandDate, PlanTaskTypeEnum.TO_SYN_MOD_CHECK)) {
                    buildAndSendMessage(astSyncTaskCtrs, acctDate);
                }
                return true;
            });
        }
        LogUtil.info(LOGGER, "五矿资产同步监控-end");
        return null;
    }
    
    /**
     *
     * @param currStandDate
     * @param planTaskTypeEnum
     * @return
     */
    private boolean checkCanNotice(Date currStandDate,PlanTaskTypeEnum planTaskTypeEnum){
        List<PlanTaskLog> tasks = planTaskLogRepo.getTasks(currStandDate, planTaskTypeEnum);
        if( CollectionUtils.isEmpty(tasks)){
            PlanTaskLog planTaskLog = new PlanTaskLog();
            planTaskLog.setExecDate(currStandDate);
            planTaskLog.setTaskType(planTaskTypeEnum);
            planTaskLogRepo.createTask(planTaskLog);
            return true;
        }
        return false;
    }

    private void buildAndSendMessage(List<AstSyncTaskCtr> astSyncTaskCtrs , Date bizDate){
        StringBuilder message = new StringBuilder();
        
        for (AstSyncTaskCtr astSyncTaskCtr : astSyncTaskCtrs) {
            AstSyncTaskCtrLog lastSyncTaskLog = astSyncTaskCtrRepo.getTaskLogByExecDate(bizDate, SyncTaskTypeEnum.AST_SYNC,
                    astSyncTaskCtr.getOrgCode());
            AstSyncTaskCtrLog lastModTaskLog = astSyncTaskCtrRepo.getTaskLogByExecDate(bizDate, SyncTaskTypeEnum.AST_MOD,
                    astSyncTaskCtr.getOrgCode());
            if (lastSyncTaskLog == null) {
                message.append(astSyncTaskCtr.getOrgCode() + ":日期为:" + DateTimeUtil.formatY_M_D(bizDate) + "数据同步未完成");
            } else {
                message.append(astSyncTaskCtr.getOrgCode() + ":日期为:" + DateTimeUtil.formatY_M_D(bizDate) + "数据同步已完成");
            }
            message.append("\n");
            if (lastModTaskLog == null) {
                message.append(astSyncTaskCtr.getOrgCode() + ":日期为:" + DateTimeUtil.formatY_M_D(bizDate) + "模型转换未完成");
            } else {
                message.append(astSyncTaskCtr.getOrgCode() + ":日期为:" + DateTimeUtil.formatY_M_D(bizDate) + "模型转换已完成");
            }
        }
        message.append("\n");
        Set<String> receivers = super.getAdminReceivers();
        PlanBase planBase=new PlanBase();
        super.sendBussinessNotice(planBase, SequenceUtil.genFakeSeqNo(), message.toString(), receivers);
    }

    @Override
    protected IEnum getBizType() {
        return AbssqrNoticeTypeEnum.WkMfsSync;
    }

    @Override
    protected String getWatchJob() {
        return ZLOriginalAstSyncJob.class.getName();
    }

}
