/**
 * abssqr.com Inc. 
 * Copyright (c) 2017-2018 All Rights Reserved. 
 */
package com.abssqr.plat.common.integration.utils;

import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

import com.abssqr.plat.common.util.loggers.CommonLoggers;
import com.google.gson.Gson;
import org.apache.http.Header;
import org.apache.http.client.HttpClient;
import org.apache.http.impl.client.DefaultConnectionKeepAliveStrategy;
import org.apache.http.impl.client.DefaultHttpRequestRetryHandler;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.apache.http.message.BasicHeader;
import org.apache.http.protocol.HTTP;
import org.slf4j.Logger;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.context.annotation.Lazy;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.client.HttpComponentsClientHttpRequestFactory;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.http.converter.StringHttpMessageConverter;
import org.springframework.stereotype.Component;
import org.springframework.web.client.DefaultResponseErrorHandler;
import org.springframework.web.client.RestTemplate;

/**
 *
 * @author zhenxuan.luo
 * @version com.abssqr.plat.common.integration.utils: RestTemplateClient.java, v 0.1 2018-01-22 下午20:56 zhenxuan.luo Exp $
 */
@Component
@Lazy(false)
public class RestTemplateClient implements InitializingBean {
    
    private final static Logger LOGGER = CommonLoggers.COMMON_HTTP;
    
    private volatile RestTemplate restTemplate;
    
    private long liveTime = 30;
    
    private int maxSize = 30;
    
    private int retryCount = 3;
    
    private int connectionRequestTimeOut = 1000;
    
    private int connectTimeOut = 15000;
    
    private int readTimeOut = 15000;
    
    public String postJson(String url, Object obj) {
        HttpHeaders headers = new HttpHeaders();
        MediaType type = MediaType.parseMediaType("application/json; charset=UTF-8");
        headers.setContentType(type);
        headers.add("Accept", MediaType.APPLICATION_JSON.toString());
        
        String json = new Gson().toJson(obj);
        HttpEntity<String> formEntity = new HttpEntity<String>(json, headers);
        LOGGER.info("invoke [" + url + "]: {" + json + "}");
        String result = restTemplate.postForObject(url, formEntity, String.class);
        return result;
    }
    public String postXml(String url, String xml) {
        HttpHeaders headers = new HttpHeaders();
        MediaType type = MediaType.parseMediaType("text/xml; charset=UTF-8");
        headers.setContentType(type);
        headers.add("Accept", MediaType.APPLICATION_XML_VALUE.toString());

        HttpEntity<String> formEntity = new HttpEntity<String>(xml, headers);
        LOGGER.info("invoke [" + url + "]: {" + xml + "}");
        String result = restTemplate.postForObject(url, formEntity, String.class);
        return result;
    }
    @Override
    public void afterPropertiesSet() {
        doInit();
    }
    
    private RestTemplate getRestTemplate() {
        if (null == restTemplate) {
            doInit();
        }
        
        return restTemplate;
    }
    
    private void doInit() {
        synchronized (this) {
            if (restTemplate != null) {
                return;
            }
        }
        
        LOGGER.info("RestTemplateClient初始化");
        
        // 长连接保持
        PoolingHttpClientConnectionManager pollingConnectionManager = new PoolingHttpClientConnectionManager(30,
                TimeUnit.SECONDS);
        // 总连接数
        pollingConnectionManager.setMaxTotal(maxSize);
        // 同路由的并发数
        pollingConnectionManager.setDefaultMaxPerRoute(maxSize);
        HttpClientBuilder httpClientBuilder = HttpClients.custom();
        httpClientBuilder.setConnectionManager(pollingConnectionManager);
        
        // 重试次数，默认是3次，没有开启
        httpClientBuilder.setRetryHandler(new DefaultHttpRequestRetryHandler(retryCount, true));
        
        // 保持长连接配置，需要在头添加Keep-Alive
        httpClientBuilder.setKeepAliveStrategy(new DefaultConnectionKeepAliveStrategy());
        
        // 请求头配置
        List<Header> headers = new ArrayList<Header>();
        headers.add(new BasicHeader("User-Agent",
                "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/31.0.1650.16 Safari/537.36"));
        headers.add(new BasicHeader("Accept-Encoding", "gzip,deflate"));
        headers.add(new BasicHeader("Accept-Language", "zh-CN"));
        headers.add(new BasicHeader("Connection", "Keep-Alive"));
        headers.add(new BasicHeader(HTTP.CONTENT_TYPE, "application/json"));
        
        httpClientBuilder.setDefaultHeaders(headers);
        
        HttpClient httpClient = httpClientBuilder.build();
        
        // httpClient连接配置，底层是配置RequestConfig
        HttpComponentsClientHttpRequestFactory clientHttpRequestFactory = new HttpComponentsClientHttpRequestFactory(
                httpClient);
        // 连接超时
        clientHttpRequestFactory.setConnectTimeout(connectTimeOut);
        // 数据读取超时时间，即SocketTimeout
        clientHttpRequestFactory.setReadTimeout(readTimeOut);
        // 连接不够用的等待时间，不宜过长，必须设置，比如连接不够用时，时间过长将是灾难性的
        clientHttpRequestFactory.setConnectionRequestTimeout(connectionRequestTimeOut);
        // 缓冲请求数据，默认值是true。通过POST或者PUT大量发送数据时，建议将此属性更改为false，以免耗尽内存。
        clientHttpRequestFactory.setBufferRequestBody(false);
        
        // 添加内容转换器
        List<HttpMessageConverter<?>> messageConverters = new ArrayList<HttpMessageConverter<?>>();
        messageConverters.add(new StringHttpMessageConverter(Charset.forName("UTF-8")));
        
        restTemplate = new RestTemplate(messageConverters);
        restTemplate.setRequestFactory(clientHttpRequestFactory);
        restTemplate.setErrorHandler(new DefaultResponseErrorHandler());
        
        LOGGER.info("RestTemplateClient初始化完成");
    }
}