package com.abssqr.plat.common.util.utils;

import com.abssqr.plat.common.util.convert.copier.BeanCopierConvert;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.springframework.cglib.beans.BeanCopier;
import org.springframework.cglib.beans.BeanMap;
import org.springframework.util.CollectionUtils;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * 相似BeanCopy工具类,减少copy编码,相同的bean初始化一次,以提升性能
 * <p/>
 *
 * @author bangis.wangdf
 * @date 16/5/14.23:44
 */
public class BeanCopierUtil {
    
    /**
     * BeanCopier 缓存Map
     **/
    private final static Map<String, BeanCopier> COPIER_MAP = Maps.newConcurrentMap();
    
    /**
     * BeanCopy
     *
     * @param targetClass 目标
     * @param source      源
     * @param <E>         E
     * @return result
     */
    public static <E> E copy(Class<E> targetClass, Object source) {
        if (source == null) {
            return null;
        }
        
        String sourceName = source.getClass().getCanonicalName();
        sourceName += targetClass.getCanonicalName();
        try {
            E result = targetClass.newInstance();
            BeanCopier beanCopier = getBeanCopier(targetClass, source.getClass(), sourceName, true);
            beanCopier.copy(source, result, BeanCopierConvert.converter());
            return result;
        } catch (Throwable e) {
            throw new RuntimeException("模型转换错误,Class:" + targetClass + ",source:" + source, e);
        }
        
    }
    
    /**
     * ListCopy
     *
     * @param targetClass 目标
     * @param sourceList  源
     * @param <E>         E
     * @return result
     */
    public static <E> List<E> copyList(Class<E> targetClass, List<?> sourceList) {
        if (CollectionUtils.isEmpty(sourceList)) {
            return new ArrayList<E>();
        }
        
        List<E> result = Lists.newArrayList();
        for (Object obj : sourceList) {
            result.add(copy(targetClass, obj));
        }
        return result;
    }
    
    /**
     * beanCopier 获取
     *
     * @param classT classT
     * @param classF classF
     * @param mapKey mapKey
     * @return BeanCopier
     */
    private static BeanCopier getBeanCopier(Class classT, Class classF, String mapKey,
                                            boolean useConverter) {
        BeanCopier beanCopier = COPIER_MAP.get(mapKey);
        if (beanCopier == null) {
            beanCopier = BeanCopier.create(classF, classT, useConverter);
            COPIER_MAP.put(mapKey, beanCopier);
        }
        return beanCopier;
    }
    
    /**
     * 将对象装换为map
     * @param bean
     * @return
     */
    public static <T> Map<String, Object> beanToMap(T bean) {
        Map<String, Object> map = Maps.newHashMap();
        if (bean != null) {
            BeanMap beanMap = BeanMap.create(bean);
            for (Object key : beanMap.keySet()) {
                map.put(key + "", beanMap.get(key));
            }
        }
        return map;
    }
    
    /**
     * 将map装换为javabean对象
     * @param map
     * @param bean
     * @return
     */
    public static <T> T mapToBean(Map<String, Object> map, T bean) {
        BeanMap beanMap = BeanMap.create(bean);
        beanMap.putAll(map);
        return bean;
    }
    
    /**
     * 将List<T>转换为List<Map<String, Object>>
     * @param objList
     * @return
     */
    public static <T> List<Map<String, Object>> objectsToMaps(List<T> objList) {
        List<Map<String, Object>> list = Lists.newArrayList();
        if (objList != null && objList.size() > 0) {
            Map<String, Object> map = null;
            T bean = null;
            for (int i = 0, size = objList.size(); i < size; i++) {
                bean = objList.get(i);
                map = beanToMap(bean);
                list.add(map);
            }
        }
        return list;
    }
    
    /**
     * 将List<Map<String,Object>>转换为List<T>
     * @param maps
     * @param clazz
     * @return
     */
    public static <T> List<T> mapsToObjects(List<Map<String, Object>> maps, Class<T> clazz)
            throws InstantiationException, IllegalAccessException {
        List<T> list = Lists.newArrayList();
        if (maps != null && maps.size() > 0) {
            Map<String, Object> map = null;
            T bean = null;
            for (int i = 0, size = maps.size(); i < size; i++) {
                map = maps.get(i);
                bean = clazz.newInstance();
                mapToBean(map, bean);
                list.add(bean);
            }
        }
        return list;
    }
    
    /**
     * 将一个 Map 对象转化为一个 JavaBean
     * @param clazz 要转化的类型
     * @param map 包含属性值的 map
     * @return 转化出来的 JavaBean 对象
     */
    @SuppressWarnings("rawtypes")
    public static <T> T toBean(Class<T> clazz, Map map) {
        T obj = null;
        try {
            BeanInfo beanInfo = Introspector.getBeanInfo(clazz);
            obj = clazz.newInstance(); // 创建 JavaBean 对象
            
            // 给 JavaBean 对象的属性赋值
            PropertyDescriptor[] propertyDescriptors = beanInfo.getPropertyDescriptors();
            for (int i = 0; i < propertyDescriptors.length; i++) {
                PropertyDescriptor descriptor = propertyDescriptors[i];
                String propertyName = descriptor.getName();
                if (map.containsKey(propertyName)) {
                    // 下面一句可以 try 起来，这样当一个属性赋值失败的时候就不会影响其他属性赋值。
                    Object value = map.get(propertyName);
                    if (value instanceof LinkedHashMap) {
                        System.out.println(value.getClass());
                        value = toBean(descriptor.getPropertyType(), (Map) value);
                    }
                    try {
                        descriptor.getWriteMethod().invoke(obj, value);
                    } catch (InvocationTargetException e) {
                        System.out.println("字段映射失败");
                    }
                }
            }
        } catch (IllegalAccessException e) {
            System.out.println("实例化 JavaBean 失败");
        } catch (IntrospectionException e) {
            System.out.println("分析类属性失败");
        } catch (IllegalArgumentException e) {
            System.out.println("映射错误");
        } catch (InstantiationException e) {
            System.out.println("实例化 JavaBean 失败");
        }
        return (T) obj;
    }
    
}
