/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2019 All Rights Reserved.
 */
package com.abssqr.plat.core.service.ast;

import com.abssqr.plat.common.facade.base.DebtMoney;
import com.abssqr.plat.common.facade.base.OrgDistPftMoney;
import com.abssqr.plat.common.facade.enums.DebtMoneyFieldEnum;
import com.abssqr.plat.common.facade.enums.EventCodeEnum;
import com.abssqr.plat.common.facade.enums.LoanStatusEnum;
import com.abssqr.plat.common.facade.enums.TransCodeEnum;
import com.abssqr.plat.common.facade.utils.CalcUtil;
import com.abssqr.plat.common.model.domain.ast.Loan;
import com.abssqr.plat.common.model.domain.ast.LoanAst;
import com.abssqr.plat.common.model.domain.ast.LoanAstProfit;
import com.abssqr.plat.common.model.domain.ast.LoanInstal;
import com.abssqr.plat.common.model.domain.ast.LoanInstalDomain;
import com.abssqr.plat.common.model.domain.astSync.BaseTsfRes;
import com.abssqr.plat.common.model.domain.astSync.LoanAstSyncRepayPlan;
import com.abssqr.plat.common.model.domain.astSync.LoanAstSyncRepayPlanTrans;
import com.abssqr.plat.common.model.domain.astSync.LoanAstSyncTrans;
import com.abssqr.plat.common.model.domain.clearing.PlanClearingLogKey;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstLog;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstProfitLog;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstProfitTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.LoanInstalLog;
import com.abssqr.plat.common.model.domain.loanTrans.LoanInstalTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.LoanTransDetail;
import com.abssqr.plat.common.util.profiler.PerformTraceContext;
import com.abssqr.plat.core.service.plan.PlanClearingManager;
import com.abssqr.plat.core.service.plan.PlanManager;
import com.abssqr.plat.core.service.repo.AstFlowOutComponent;
import com.abssqr.plat.core.share.context.PlanContextHolder;
import com.general.system.common.model.Money;
import com.general.system.common.model.exception.BaseCommonException;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.MessageUtil;
import com.general.system.common.util.MoneyUtil;
import com.general.system.common.util.VarChecker;
import com.google.common.collect.Maps;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * 回收处理 (还款)
 *
 * @author zhenxuan.luo
 * @version LoanAstRepoManager.java, v 0.1 2019-06-17 20:07 zhenxuan.luo Exp $
 */
@Component
public class LoanAstRepoManager extends LoanAstBaseManager {
    
    @Autowired
    private PlanClearingManager planClearingManager;
    @Autowired
    private LoanStatusManager   loanStatusManager;
    @Autowired
    private PlanManager         planManager;
    @Autowired
    private AstFlowOutComponent astFlowOutComponent;
    
    @Override
    protected Loan doProcess(BaseTsfRes baseTsfRes, LoanAstSyncTrans loanAstSyncTrans,
                             LoanTransDetail loanTransDetail) {
        
        Loan loan = loanTransDetail.getLoan();
        Map<Integer, LoanInstalDomain> loanInstalMap = loan.getLoanInstalMap();
        Map<Integer, LoanAstSyncRepayPlan> syncRepayPlanMap = baseTsfRes.getRepayPlans().stream().collect(
                Collectors.toMap(LoanAstSyncRepayPlan::getTermNo, item -> item));
        
        try {
            PerformTraceContext.enter("LoanAstRepoManager.allRepayPlan");
            for (LoanAstSyncRepayPlanTrans repayPlanTrans : loanAstSyncTrans.getRepayPlanTransList()) {
                
                LoanInstalDomain refInstal = loanInstalMap.get(repayPlanTrans.getTermNo());
                
                EventCodeEnum eventCode;
                if (refInstal.getStatus() == LoanStatusEnum.OVERDUE) {
                    // 分期已逾期：逾期还款
                    eventCode = EventCodeEnum.ODUE_REPAY;
                } else if (DateTimeUtil.calcDayDiff(loanAstSyncTrans.getAccountDate(),
                        refInstal.getAstPeriod().getEndDate()) < 0) {
                    // 还款会计日期早于分期到期日：提前还款
                    eventCode = EventCodeEnum.PRE_REPAY;
                } else {
                    eventCode = EventCodeEnum.DUE_REPAY;
                }
                LoanInstalTransDetail loanInstalTransDetail = loanTransDetail
                                                                      .addLoanInstalTransDtl(refInstal, eventCode);
                loanInstalTransDetail.getActiveLog().setTransTime(repayPlanTrans.getTransTime());
                loanInstalTransDetail.getActiveLog().setOutBizNo(repayPlanTrans.getOutBizNo());
                
                // 1. 处理金额
                loanInstalTransDetail.processTransAmt(repayPlanTrans.getAmt());
                loanInstalTransDetail.processPaidFundAmt(repayPlanTrans.getFundAmt());
                astFlowOutComponent
                        .flowOutInstal(loanInstalTransDetail, repayPlanTrans.getAmt(), repayPlanTrans.getFundAmt(),
                                false);

                // 2. 处理结清
                LoanAstSyncRepayPlan syncRepayPlan = syncRepayPlanMap.get(repayPlanTrans.getTermNo());
                if (isNeedSetClear(syncRepayPlan.getStatus(), syncRepayPlan.getAstPeriod().getClearDate(), refInstal,
                        loanTransDetail.getActiveLog().getAccountDate())) {
                    refInstal.getBal().checkZero(MessageUtil.formatMsg("结清分期[{0}-{1}]的余额[{2}]必须为0. ",
                            refInstal.getAstNo(), refInstal.getTermNo(), refInstal.getBal()));
                    
                    Date clearDate = syncRepayPlan.getAstPeriod().getClearDate();
                    if (repayPlanTrans.getTermNo() == 0) {
                        clearDate = loanTransDetail.getActiveLog().getAccountDate();
                    }
                    loanInstalTransDetail.setClear(clearDate);
                }
                
                loanInstalTransDetail.setClrgAmt(repayPlanTrans.getFundAmt().getSum());
                
                // 4. 余额校验
                refInstal.getBal().check(MessageUtil.formatMsg("分期[{0}-{1}]余额因[{2}]导致非法", refInstal.getAstNo(),
                        refInstal.getTermNo(), repayPlanTrans));
            }
        }catch (BaseCommonException e){
            LogUtil.error(LOGGER, e, "LoanAstRepoManager执行还款异常：" + e.getMessage());
            throw e;
        }finally {
            PerformTraceContext.leave();
            PlanContextHolder.clear();
        }
        
        // 还款清算
        planClearingManager.processClearingForLoanAst(loanTransDetail.getLoanInstalTransDetails(), TransCodeEnum.REPO,
                loanAstSyncTrans.getAccountDate(), instalTransDetail -> {
                    Map<PlanClearingLogKey, Money> ret = Maps.newHashMap();
                    // 根据机构资产持有收益累加还款清算金额
                    buildLoanAstClearing(instalTransDetail, ret);
                    // 根据收益累加还款清算金额
                    buildLoanAstProfitClearing(instalTransDetail, ret);
                    return ret;
                });
        
        // 应付利息的清算
        planClearingManager.processClearing(loanTransDetail.getLoanInstalTransDetails(), TransCodeEnum.PAY_HANG_INTR,
                loanAstSyncTrans.getAccountDate(),
                instalTransDetail -> {
                    Money hangClrAmt = new Money();
                    Money hangIntrAmt = new Money();
                    Money allPaidIntAndFee = new Money();
                    Money allPaidFundIntAndFee = new Money();
                    //计算金融服务费-摊出金额
                    Money fsApporAmt = new Money();
                    LoanInstal refInstal = instalTransDetail.getLoanInstal();
                    
                    //判读是否为兼容模式
                    //根据分期上还款获取计算分摊比例金额 （已验证（20191030上线前）转让后有挂账，转让前无还款记录）
                    allPaidIntAndFee = refInstal.getPaidAmt().cloneAll(DebtMoneyFieldEnum.OTHER_FEE).getAllIntAndFee();
                    allPaidFundIntAndFee = refInstal.getPaidFundAmt().cloneAll(DebtMoneyFieldEnum.OTHER_FEE).getAllIntAndFee();
                    //获取待处理的挂账金额
                    hangIntrAmt = MoneyUtil.min(refInstal.getHangIntrBal(), allPaidIntAndFee);
                    
                    if (refInstal.getStatus() == LoanStatusEnum.CLEAR) {
                        hangIntrAmt = refInstal.getHangIntrBal();
                    }
                    if (hangIntrAmt.isGreaterThanZero()) {
                        //计算本次清算金额
                        hangClrAmt = CalcUtil.calcMoneyFactor(hangIntrAmt, allPaidFundIntAndFee.getAmount(),
                                allPaidIntAndFee.getAmount());
                        //计算金融服务费-摊出金额: = 获取待处理的挂账金额-计算本次清算金额
                        fsApporAmt = hangIntrAmt.subtract(hangClrAmt);
                    }
                    
                    //本次处理的挂账金额以及清算金额
                    instalTransDetail.dealHangIntrAmt(hangIntrAmt, hangClrAmt);
                    //处理金融服务费
                    instalTransDetail.calcOrgFeeSplit(hangClrAmt, fsApporAmt);
                    checkOrgFeeSplit(instalTransDetail);
                    return hangClrAmt;
                });
        
        // 资产状态与五级分类变更
        loanStatusManager.calcLoanRiskStatus(loanTransDetail, false);
        
        return loanTransDetail.getLoan();
    }
    
    private void checkOrgFeeSplit(LoanInstalTransDetail transDetail) {
        LoanInstalLog activeLog = transDetail.getActiveLog();
        //本次处理挂账金额
        Money hangIntrAmt = activeLog.getHangIntrAmt().clone();
        //本次实付挂账金额
        Money paidHangIntrFundAmt = activeLog.getPaidHangIntrFundAmt().clone();
        
        //金融服务费
        OrgDistPftMoney orgDistPftAmt = activeLog.getOrgDistPftAmt();
        //应收
        DebtMoney payAmt = activeLog.getAmt().cloneAll();
        //实付
        DebtMoney paidFundAmt = activeLog.getPaidFundAmt().cloneAll();
        
        VarChecker.checkArgument(paidFundAmt.getSum().add(orgDistPftAmt.getAmt()).equals(payAmt.getSum()),
                "机构分润校验失败：应还=实还+分润：[{0}]", activeLog);
        VarChecker.checkArgument(paidFundAmt.getSum().add(orgDistPftAmt.getApporAmt()).add(orgDistPftAmt.getPaidAmt())
                                         .equals(payAmt.getSum()),
                "机构分润校验摊出失败：应还=实还+分润(摊出金额+机构实付分润)：[{0}]", activeLog);
        VarChecker.checkArgument(hangIntrAmt.equals(paidHangIntrFundAmt.add(orgDistPftAmt.getApporAmt())),
                "机构分润校验摊出失败：挂账金额=挂账清算金额+摊出金额：[{0}]", activeLog);
    }
    
    private boolean isNeedSetClear(LoanStatusEnum syncStatus, Date syncClearDate, LoanInstal instal, Date accountDate) {
        
        if (syncStatus == LoanStatusEnum.CLEAR || syncClearDate != null) {
            return true;
        }
        
        // 资产未到期，不设置结清
        if (DateTimeUtil.calcDayDiff(accountDate, instal.getAstPeriod().getEndDate()) < 0) {
            return false;
        }
        
        if (instal.getBal().getSum().isZero()) {
            return true;
        }
        
        return false;
    }
    
    private void buildLoanAstClearing(LoanInstalTransDetail instalTransDetail,
                                      Map<PlanClearingLogKey, Money> clearingMap) {
        for (LoanAstTransDetail astTransDetail : instalTransDetail.getLoanAstTransDetails()) {
            LoanAstLog loanAstLog = astTransDetail.getActiveLog();
            PlanClearingLogKey key = new PlanClearingLogKey(loanAstLog.getOwnOrg(), loanAstLog.getLenderOrgCode(),
                    loanAstLog.getEventCode().getSceneCode());
            
            if (!clearingMap.containsKey(key)) {
                clearingMap.put(key, new Money());
            }
            clearingMap.get(key).addTo(loanAstLog.getClrgAmt());
        }
    }
    
    private void buildLoanAstProfitClearing(LoanInstalTransDetail instalTransDetail,
                                            Map<PlanClearingLogKey, Money> clearingMap) {
        for (LoanAstProfitTransDetail astProfitTransDetail : instalTransDetail.getLoanAstProfitTransDetails()) {
            LoanAstProfitLog loanAstProfitLog = astProfitTransDetail.getActiveLog();
            LoanAstProfit loanAstProfit = astProfitTransDetail.getLoanAstProfit();
            LoanAst loanAst = loanAstProfit.getRefLoanAst();
            PlanClearingLogKey key = new PlanClearingLogKey(loanAstProfitLog.getOwnOrg(), loanAst.getLenderOrgCode(),
                    loanAstProfitLog.getEventCode().getSceneCode());
            
            if (!clearingMap.containsKey(key)) {
                clearingMap.put(key, new Money());
            }
            clearingMap.get(key).addTo(loanAstProfitLog.getClrgAmt());
        }
    }
}