/**
 * abssqr.com Inc. 
 * Copyright (c) 2017-2019 All Rights Reserved. 
 */
package com.abssqr.plat.core.service.plan;

import cn.hutool.core.date.DateUtil;
import cn.hutool.core.util.StrUtil;
import com.abssqr.plat.common.dal.mysql.auto.resultmap.PlanStatItem;
import com.abssqr.plat.common.facade.enums.AstFlowTypeEnum;
import com.abssqr.plat.common.facade.enums.PlanCapitalTransTypeEnum;
import com.abssqr.plat.common.facade.enums.PlanTypeEnum;
import com.abssqr.plat.common.facade.enums.StageChangeTypeEnum;
import com.abssqr.plat.common.facade.enums.TaskStatusEnum;
import com.abssqr.plat.common.facade.enums.TransCodeEnum;
import com.abssqr.plat.common.facade.model.plan.PlanAlloc;
import com.abssqr.plat.common.facade.model.plan.PlanAst;
import com.abssqr.plat.common.facade.model.trans.CapitalTrader;
import com.abssqr.plat.common.facade.model.trans.ChangeStageParam;
import com.abssqr.plat.common.model.domain.plan.PlanAllocFrozen;
import com.abssqr.plat.common.model.domain.plan.PlanAstLog;
import com.abssqr.plat.common.model.enums.TraderTypeEnum;
import com.abssqr.plat.common.model.repo.plan.PlanAllocFrozenRepo;
import com.abssqr.plat.common.model.repo.plan.PlanAstLogRepository;
import com.abssqr.plat.common.model.repo.plan.PlanAstRepository;
import com.abssqr.plat.common.model.utils.PlanAstChangeCheckerUtil;
import com.general.system.common.model.Money;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.VarChecker;
import com.general.system.tool.util.ToolUtil;
import com.google.common.collect.Lists;
import org.apache.commons.collections.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * 计划资产余额水位管理
 * @author hanfei
 * @version $Id: PlanAstManager.java, v 0.1 2019-07-04 12:31 PM hanfei Exp $ 
 */
@Component
public class PlanAstManager {
    
    protected final Logger LOGGER = LoggerFactory.getLogger(getClass());
    
    private static ThreadLocal<PlanAst> planAstLockHolder = new ThreadLocal<>();
    @Autowired
    private PlanAstLogRepository planAstLogRepository;
    @Autowired
    private PlanAstRepository    planAstRepository;
    @Autowired
    private PlanAllocFrozenRepo  planAllocFrozenRepo;
    
    
    /**
     * 资产水位变更操作
     * @param accountDate   会计日
     * @param planNo        计划编号
     * @param extend   扩展方法
     */
    public void changePlanAst(Date accountDate, String planNo, ChangePlanAstExtend extend) {
        // 锁定计划资产水位表
        PlanAst planAst = planAstRepository.lockByPlanNo(planNo);
        VarChecker.checkNotNull(planAst, "计划编号[{0}]查询加锁资金水位失败", planNo);
        // 变更流水
        List<PlanAstLog> planAstLogs =  new ArrayList<>();
        try {
            // 执行外部业务逻辑
            extend.handle(planAst, planAstLogs);
            
            // 如果外部没有记录日志 直接退出
            if (ToolUtil.isEmpty(planAstLogs)) {
                return;
            }
            // 最近变更日期
            planAst.setLastUpdateDate(accountDate);
            //LogUtil.info(LOGGER, "{0}更新计划水位信息{1}", planAst.getPlanNo(), planAst.getCash());
            // 校验水位金额
            PlanAstChangeCheckerUtil.me().checkPlanAstAmtZero(planAst);
            // 更新水位
            planAstRepository.update(planAst);
            
            // 通过本地线程获取水位并更新
            PlanAst cntPtyPlanAst = this.getPlanAstLockHolder();
            if (cntPtyPlanAst != null
                && !StrUtil.equals(cntPtyPlanAst.getPlanNo(), planAst.getPlanNo())) {
                // 更新水位 最近变更日期
                cntPtyPlanAst.setLastUpdateDate(accountDate);
                // 校验水位金额
                PlanAstChangeCheckerUtil.me().checkPlanAstAmtZero(cntPtyPlanAst);
                planAstRepository.update(cntPtyPlanAst);
            }
            // 校验变更流水
            PlanAstChangeCheckerUtil.me().checkChangeLogAmt(planAstLogs);
            // 记录变更记录
            planAstLogRepository.createBatch(planAstLogs);
            // LogUtil.info(LOGGER, "创建变更水位信息流水{0}条", planAstLogs.size());
        }finally {
            // 清除本地缓存
            this.removePlanAstLockHolder();
        }
    }
    
    /**
     * 根据交易单最后执行水位变化
     * @param planAst
     * @param param
     * @return
     */
    public List<PlanAstLog> execChangeStageTwoWay(PlanAst planAst, ChangeStageParam param) {
        List<PlanAstLog> astLogs = new ArrayList<>();
        CapitalTrader trader = param.getCapitalTrader();
        TransCodeEnum transCode = trader.getTransCode();
    
        // 获取交易单据类型
        PlanCapitalTransTypeEnum transType = trader.getCapitalTransType();
        
        // 1、发起方变更水位 登记流水
        astLogs.add(this.itemChangeStage(param.getAccountDate(), planAst, TraderTypeEnum.INITIATOR, param.getChangeType(),
                param.getTransAmt(), param.getClearToWayAmt(), transType, param.getTransBizNo(), param.getExtInf()));
        /**
         * 判断是否操作对手方水位金额
         */
        if (!trader.isChangeCntPty()) {
            return astLogs;
        }
        // 2、通过交易对手 获取对应水位 进行变化 通过本地线程 获取
        PlanAst cntPtyPlanAst = this.getOtherPlanAstByLocal(trader.getCntPtyCode());
        // 回购双方为承接池时 需要转换交易编码
        TransCodeEnum cntPtyTransCode = (transCode == TransCodeEnum.BUY_BACK_OUT ? TransCodeEnum.BUY_BACK_IN : transCode);
        // 双边时 发起方+  对手方-  || 发起方-  对手方+
        StageChangeTypeEnum changeType = param.getChangeType();
        // 反转资金流向
        AstFlowTypeEnum changeFlow = transType.getFlowType().reverse();
        // 反转变化类型
        if (changeType.isChangeStage()){
            // 如果是变化水位金额的处理 ADD 或 SUB 需要做反向操作
            changeType = StageChangeTypeEnum.getChangeByFlow(changeFlow);
        }
        
        // 初始化流水
        PlanAstLog astLog = PlanAstLog.buildAstLog(param.getAccountDate(), cntPtyPlanAst,
                param.getTransBizNo(), TraderTypeEnum.COUNTERPARTY, cntPtyTransCode, changeType);
        // 根据 TransCodeEnum 与 StageChangeTypeEnum 变化水位及流水
        astLog.changeStageAndLog(cntPtyPlanAst, changeFlow, param.getTransAmt(), param.getClearToWayAmt(), param.getExtInf());
        LogUtil.info(LOGGER, "{0}变更水位-TransCode[{1}]-StageChange[{2}]-FLOW[{3}]-AMT[{4}]",
                cntPtyPlanAst.getPlanNo(), cntPtyTransCode.getCode(), changeType.getCode(), changeFlow.getCode(), param.getTransAmt());
        astLogs.add(astLog);
        return astLogs;
    }
    
    /**
     * 变化水位登记流水
     * @param accountDate
     * @param planAst
     * @param traderType
     * @param changeType
     * @param transAmt
     * @param transType
     * @param transBizNo
     * @param extInf
     * @return
     */
    public PlanAstLog itemChangeStage(Date accountDate, PlanAst planAst, TraderTypeEnum traderType,
                                      StageChangeTypeEnum changeType, Money transAmt,
                                      PlanCapitalTransTypeEnum transType,
                                      String transBizNo, String extInf) {
        return itemChangeStage( accountDate, planAst, traderType,changeType, transAmt, new Money(),
                transType, transBizNo, extInf);
    }
    
    /**
     * 对于清算时产生金额不一致
     * @param accountDate
     * @param planAst
     * @param traderType
     * @param changeType
     * @param transAmt
     * @param clearToWayAmt
     * @param transType
     * @param transBizNo
     * @param extInf
     * @return
     */
    private PlanAstLog itemChangeStage(Date accountDate, PlanAst planAst, TraderTypeEnum traderType,
                                      StageChangeTypeEnum changeType, Money transAmt, Money clearToWayAmt,
                                      PlanCapitalTransTypeEnum transType,
                                      String transBizNo, String extInf) {
        TransCodeEnum transCode = transType.getTransCode();
        AstFlowTypeEnum changeFlow = transType.getFlowType();
        // 初始化流水
        PlanAstLog selfAstLog = PlanAstLog.buildAstLog(
                accountDate, planAst, transBizNo, traderType, transCode, changeType);
        // 根据 TransCodeEnum 与 StageChangeTypeEnum 变化水位及流水
        selfAstLog.changeStageAndLog(planAst, changeFlow, transAmt, clearToWayAmt, extInf);
        LogUtil.info(LOGGER, "{0}变更水位-TransCode[{1}]-StageChange[{2}]-FLOW[{3}]-AMT[{4}]",
                planAst.getPlanNo(), transCode.getCode(), changeType.getCode(), changeFlow.getCode(), transAmt.getAmount());
        return selfAstLog;
    }
    
    /**
     * 资产水位冻结金额变更操作
     * 目前只面向 分配兑付
     * @param accountDate 会计日
     * @param planNo 计划编号
     * @param planAllocList 分配计划
     */
    public PlanAstLog calcFrozen(Date accountDate, String planNo, List<PlanAlloc> planAllocList, String extInf) {
        // 锁定计划资产水位表
        PlanAst planAst = planAstRepository.lockByPlanNo(planNo);
        VarChecker.checkNotNull(planAst, "计划编号[{0}]查询加锁资金水位失败", planNo);
        // 当前实际可用余额
        Money ableCash = planAst.containClearAbleBal();
        // 最终实际冻结金额合计
        Money realFrozenAmt = new Money();
        planAllocList.forEach(alloc -> {
            // 使用当前实际可用余额 计算每个分配计划需要冻结的资金  然后累计冻结金额
            realFrozenAmt.addTo(this.allocFrozenDetail(accountDate, alloc, ableCash));
        });
        
        // 冻结资金日志
        PlanAstLog planAstLog = PlanAstLog.buildAstLog(accountDate, planAst, planNo, TraderTypeEnum.INITIATOR,
                TransCodeEnum.FROZEN, StageChangeTypeEnum.FROZEN);
        // 记录冻结发生额
        planAstLog.changeFrozenCash(planAst, realFrozenAmt, realFrozenAmt, extInf);
        if (!realFrozenAmt.isGreaterThanZero()) {
            return planAstLog;
        }
        // 最近变更日期
        planAst.setLastUpdateDate(accountDate);
        // 校验水位金额
        PlanAstChangeCheckerUtil.me().checkPlanAstAmtZero(planAst);
        // 更新
        planAstRepository.update(planAst);
        // 校验变更流水
        PlanAstChangeCheckerUtil.me().checkChangeLogAmt(Lists.newArrayList(planAstLog));
        // 记录日志
        planAstLogRepository.create(planAstLog);
    
        LogUtil.info(LOGGER, "计划[{0}]在{1}冻结金额[{2}]", planAst.getPlanNo(), accountDate, realFrozenAmt.getAmount());
        return planAstLog;
    }
    
    /**
     * 获取 计算 时间段内 的日均还款 变更水位的还款金额
     * @param planNo
     * @return
     */
    public Money getRepayTransAmtStat(String planNo, Date startDate, Date endDate, boolean avg) {
        List<PlanStatItem> statItems = planAstLogRepository.getStatByTransCode(
                endDate, startDate, planNo, TransCodeEnum.REPO, StageChangeTypeEnum.ADD);
        return getAvgRepoAmt(statItems, avg);
    }
    
    /**
     * 获取 合计金额 支持 平均值
     * @param planStatItems
     * @param avg
     * @return
     */
    private Money getAvgRepoAmt(List<PlanStatItem> planStatItems, boolean avg) {
        if (CollectionUtils.isEmpty(planStatItems)) {
            return new Money();
        }
        
        Money total = new Money();
        for (PlanStatItem planStatItem : planStatItems) {
            total = total.add(planStatItem.getCashAmt());
        }
        if (avg) {
            return total.divide(planStatItems.size());
        } else {
            return total;
        }
    }
    
    /**
     * 计算当前可冻结金额
     * @param curDate   日期
     * @param alloc     分配计划
     * @param ableCash  可用水位金额
     * @return
     */
    private Money allocFrozenDetail(Date curDate, PlanAlloc alloc, Money ableCash) {
        if (ToolUtil.isOneEmpty(alloc.getAllocNo(), alloc.getPlanNo())) {
            return new Money();
        }
        // 1. 取冻结记录表
        PlanAllocFrozen allocFrozen = planAllocFrozenRepo.getByAllocNo(alloc.getPlanNo(), alloc.getAllocNo());
        // 判断当前会计日 与 冻结记录最后更新时间 小于等于 则说明已经进行了冻结
        if (allocFrozen != null && curDate.compareTo(allocFrozen.getLastUpdateDate()) <= 0){
            return new Money();
        }
        boolean ins = false;
        if (allocFrozen == null) {
            allocFrozen = PlanAllocFrozen.createNewFrozen(curDate, alloc);
            ins = true;
        }
        
        // 2. 待冻结金额 = 待分配金额 - 已冻结金额
        Money frozenAmt = alloc.calcTotalTrialAmt().subtract(allocFrozen.getFrozenCash());
        if (!frozenAmt.isGreaterThanZero()){
            return new Money();
        }
        // 3. 今日冻结金额 = 待冻结金额 / 剩余天数
        long days = DateUtil.betweenDay(curDate, alloc.getPayDate(), true);
        if (days > 3){
            // 除以天数
            frozenAmt.divideBy(days);
        }
        
        if (frozenAmt.greaterThan(ableCash)) {
            // 当前水位余额不够冻结金额 直接用当前水位金额
            LogUtil.warn(LOGGER, "分配计划[{0}]在[{1}]期望冻结金额[{2}]-实际冻结[{3}]", alloc, curDate, frozenAmt, ableCash);
            // 这里不对可用余额进行验证，有可能存在可用余额<冻结金额的情况，也按需要冻结的金额操作而不关心可用余额
            // 因为冻结金额就是个数字，是在计算可用余额的时候实际消费的，如果冻结金额>可用余额，只是控制了计划当前没有可用余额而已。
        }
        
        ableCash.subtractFrom(frozenAmt);
        
        allocFrozen.setLastFrozenCash(allocFrozen.getFrozenCash().clone());
        allocFrozen.getFrozenCash().addTo(frozenAmt);
        allocFrozen.setLastUpdateDate(curDate);
        // 4. 保存冻结记录表
        if (ins) {
            planAllocFrozenRepo.insert(allocFrozen);
        } else {
            planAllocFrozenRepo.update(allocFrozen);
        }
        return frozenAmt;
    }
    
    /**
     * 获取计划水位
     * @param planNo
     * @return
     */
    public PlanAst getByPlanNo(String planNo) {
        return planAstRepository.getByPlanNo(planNo);
    }
    
    /**
     * 根据计划-初始化科目信息
     * @param planNo
     */
    public void initPlanAst(String planNo, PlanTypeEnum entityType, Date accountDate) {
        PlanAst planAst = planAstRepository.getByPlanNo(planNo);
        if (planAst == null) {
            // 创建水位控制信息
            planAst = new PlanAst();
            planAst.setLastUpdateDate(accountDate);
            planAst.setPlanNo(planNo);
            planAst.setPlanType(entityType);
            planAst.setStatus(TaskStatusEnum.INIT);
            planAstRepository.create(planAst);
        }
        
    }
    
    /**
     * 获取对手方的水位
     * @param planNo
     * @return
     */
    private PlanAst getOtherPlanAstByLocal(String planNo) {
        if (getPlanAstLockHolder() == null) {
            setPlanAstLockHolder(planNo);
        }
        return getPlanAstLockHolder();
    }
    
    /**
     * 线程锁获取水位
     * @return
     */
    private PlanAst getPlanAstLockHolder() {
        return planAstLockHolder.get();
    }
    
    /**
     * 设置当前线程水位
     */
    private void setPlanAstLockHolder(String planNo) {
        PlanAst otherPlanAst = planAstRepository.lockByPlanNo(planNo);
        VarChecker.checkNotNull(otherPlanAst, "计划编号[{0}]查询加锁资金水位失败", planNo);
        planAstLockHolder.set(otherPlanAst);
    }
    
    /**
     * 清除缓存
     */
    private void removePlanAstLockHolder() {
        if (ToolUtil.isNotEmpty(planAstLockHolder.get())){
            planAstLockHolder.remove();
        }
    }
    
    /**
     * 根据放款池计划项目编号进行还款金额 增加水位操作
     * @param spvRepayAmtMap
     */
    public void syncRepayAddCash(Date accountDate, Map<String, BigDecimal> spvRepayAmtMap) {
        if (ToolUtil.isEmpty(spvRepayAmtMap)) {
            return;
        }
        PlanCapitalTransTypeEnum typeEnum = PlanCapitalTransTypeEnum.AST_REPO;
        // 根据 计划编号 还款金额追加水位
        for (Map.Entry<String, BigDecimal> spvRepayAmt : spvRepayAmtMap.entrySet()) {
            Money repayAmt = new Money(spvRepayAmt.getValue());
            if (!repayAmt.isGreaterThanZero()) {
                continue;
            }
            this.changePlanAst(accountDate, spvRepayAmt.getKey(), (planAst,planAstLogs)  -> {
                // 放款池同步完成追加还款金额 需要先增加校验金额 满足后续变更水位校验
                this.changeCheckCashAmt(planAst, repayAmt, typeEnum.getFlowType());
                // 还款增加在途
                planAstLogs.add(itemChangeStage(accountDate, planAst, TraderTypeEnum.INITIATOR,
                        StageChangeTypeEnum.ON_WAY, repayAmt,
                        typeEnum,
                        spvRepayAmt.getKey(), "放款池还款增加在途金额"));
                // 还款增加水位
                planAstLogs.add(itemChangeStage(accountDate, planAst, TraderTypeEnum.INITIATOR,
                        StageChangeTypeEnum.ADD, repayAmt,
                        typeEnum,
                        spvRepayAmt.getKey(), "放款池还款增加水位"));
            });
        }
    }
    
    /**
     * 针对特定场景操作校验水位金额
     * 1、放款池放款终止转回款户 需要先增加校验金额 满足后续变更水位校验
     * 2、根据放款池计划项目编号进行还款金额 增加水位操作
     * 3、实际支付金额小于应付金额 将差额追加到校验金额上满足后续变更水位校验
     * @param planAst  水位
     * @param happenAmt  发生额
     * @param flowType  资金流向
     */
    public void changeCheckCashAmt(PlanAst planAst, Money happenAmt, AstFlowTypeEnum flowType) {
        switch (flowType) {
            case FLOW_IN:
                planAst.getCheckCash().addTo(happenAmt.clone());
                break;
            case FLOW_OUT:
                planAst.getCheckCash().subtractFrom(happenAmt.clone());
                break;
            default:
                // 无变化
        }
    }
    
    /**
     * 释放冻结资金 扩展接口
     */
    public interface ChangePlanAstExtend {
        /**
         * 执行扩展
         * @param planAst
         * @return
         */
        void handle(PlanAst planAst, List<PlanAstLog> planAstLogs);
    }
}
