/**
 * abssqr.com Inc. 
 * Copyright (c) 2017-2019 All Rights Reserved. 
 */
package com.abssqr.plat.core.service.profit;

import com.abssqr.plat.common.facade.base.DebtMoney;
import com.abssqr.plat.common.facade.enums.DebtMoneyFieldEnum;
import com.abssqr.plat.common.facade.enums.EventCodeEnum;
import com.abssqr.plat.common.facade.enums.PriceTypeEnum;
import com.abssqr.plat.common.facade.enums.ProfitTypeEnum;
import com.abssqr.plat.common.facade.enums.TransCodeEnum;
import com.abssqr.plat.common.facade.model.plan.PlanAstPriceRule;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.model.domain.ast.AstProfitUnit;
import com.abssqr.plat.common.model.domain.ast.InstalAstMeasure;
import com.abssqr.plat.common.model.domain.ast.Loan;
import com.abssqr.plat.common.model.domain.ast.LoanAst;
import com.abssqr.plat.common.model.domain.ast.LoanAstGroup;
import com.abssqr.plat.common.model.domain.ast.LoanAstMeasure;
import com.abssqr.plat.common.model.domain.ast.LoanAstProfit;
import com.abssqr.plat.common.model.domain.ast.LoanInstalDomain;
import com.abssqr.plat.common.model.domain.ast.ProfitMeasure;
import com.abssqr.plat.common.model.domain.loanTrans.LoanInstalTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.LoanTransDetail;
import com.abssqr.plat.common.model.utils.AstPriceUtil;
import com.abssqr.plat.common.model.utils.LoanAstUtil;
import com.abssqr.plat.core.service.plan.PlanManager;
import com.abssqr.plat.core.share.context.PlanContextHolder;
import com.general.system.common.model.Money;
import com.general.system.common.util.VarChecker;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * 收益拆分组件
 * @author zhenxuan.luo
 * @version com.abssqr.plat.core.service.profit: ProfitSplitManager.java, v 0.1 2019-11-17 下午22:19 zhenxuan.luo Exp $ 
 */
@Component
public class ProfitSplitComponent {
    
    private final static Logger LOGGER = LoggerFactory.getLogger(ProfitSplitComponent.class);
    
    private final static List<DebtMoneyFieldEnum> INSTAL_SPLIT_FIELDS = Lists.newArrayList();
    
    private static List<TransCodeEnum> allowedTransCodes =
            Lists.newArrayList(TransCodeEnum.INTR_FEE_SETTLEMENT, TransCodeEnum.AST_SIMU,TransCodeEnum.HANG_FEE);
    
    static {
        INSTAL_SPLIT_FIELDS.add(DebtMoneyFieldEnum.INTR);
        INSTAL_SPLIT_FIELDS.add(DebtMoneyFieldEnum.PRIN_ODUE_PNY);
        INSTAL_SPLIT_FIELDS.add(DebtMoneyFieldEnum.INTR_ODUE_PNY);
        INSTAL_SPLIT_FIELDS.add(DebtMoneyFieldEnum.FEE);
        INSTAL_SPLIT_FIELDS.add(DebtMoneyFieldEnum.FEE_PNY);
        INSTAL_SPLIT_FIELDS.add(DebtMoneyFieldEnum.OTHER_FEE);
        //INSTAL_SPLIT_FIELDS.add(DebtMoneyFieldEnum.OTHER_FEE_PNY);
    }
    
    @Autowired
    private ProfitSplitCalculator profitSplitCalculator;
    
    @Autowired
    private PlanManager planManager;
    
    /**
     * 执行交易 拆分利息，本金罚息，利息罚息，费用，费用罚息
     * @param loanTrans     根据交易明细
     * @param instalTrans   分期交易明细
     * @param transAmt  交易金额
     */
    public void splitInstal(LoanTransDetail loanTrans, LoanInstalTransDetail instalTrans, DebtMoney transAmt) {
        doSplitInstal(loanTrans, instalTrans, transAmt, INSTAL_SPLIT_FIELDS,Maps.newHashMap(), "");
    }

    /**
     * 执行交易 拆分提前还款服务费(组装ProfitUnit Base=提前还款服务费发生额)
     * @param loanTrans     根据交易明细
     * @param instalTrans   分期交易明细
     * @param transAmt  交易金额
     */
    public void splitOtherFeeInstal(LoanTransDetail loanTrans, LoanInstalTransDetail instalTrans, DebtMoney transAmt,Map<String,Money> loanAstPrinAmtMap, String memo) {
        doSplitInstal(loanTrans, instalTrans, transAmt, INSTAL_SPLIT_FIELDS,loanAstPrinAmtMap, memo);
    }

    /**
     *
     * @param loanTrans
     * @param instalTrans
     * @param transAmt
     * @param fields
     * @param loanAstPrinAmtMap 由于机构资产本金发生额引起的分摊
     */
    private void doSplitInstal(LoanTransDetail loanTrans, LoanInstalTransDetail instalTrans, DebtMoney transAmt,
                             List<DebtMoneyFieldEnum> fields,Map<String,Money> loanAstPrinAmtMap, String memo) {
        if (transAmt.isZero()) {
            return;
        }
        // 结息相关事件码准入
        EventCodeEnum eventCode = instalTrans.getActiveLog().getEventCode();
        VarChecker.checkArgument(allowedTransCodes.contains(eventCode.getSceneCode()),
                "只允许结息交易进行拆分。传入事件为[{0}]",
                instalTrans.getActiveLog());
        // 贷款资产
        Loan loan = loanTrans.getLoan();
        // 分期
        LoanInstalDomain instalDomain = instalTrans.getLoanInstal();
        // 提取未来期的机构资产
        LoanAstGroup astGroup = loan.filterLaterLoanAstByTermNo(instalDomain.getTermNo());
        Date accountDate = loanTrans.getActiveLog().getAccountDate();
        //
        Map<Integer, InstalAstMeasure> instalMeasureMap = buildInstalMeasure(instalDomain, astGroup,
                accountDate,loanAstPrinAmtMap);
        
        /**
         * 处理息费发生
         * 1. 先将息费根据分配项"AstProfitUnit::base"分配到未来的每一个分期
         * 2. 考虑到每个分期下面可能有多个机构资产，因此需要将分期的分配结果摊分到机构资产
         * 3. 每个机构资产下，对应srcTerm会有一个AstProfitGroup。对于已转资产，还需要对它处理，将持有收益进行分润
         */
        // 1. 先将息费根据分配项"AstProfitUnit::base"分配到未来的每一个分期
        for (DebtMoneyFieldEnum debtMoneyField : fields) {
            Map<Integer, AstProfitUnit> profitUnit = extractUnitForInstal(instalMeasureMap, debtMoneyField);
            profitSplitCalculator.splitDebtMoney(transAmt.getByFieldType(debtMoneyField), profitUnit);
        }
        
        // 2. 考虑到每个分期下面可能有多个机构资产，因此需要将分期的分配结果摊分到机构资产
        for (Integer termNo : instalMeasureMap.keySet()) {
            InstalAstMeasure instalMeasure = instalMeasureMap.get(termNo);
            
            for (DebtMoneyFieldEnum debtMoneyField : fields) {
                AstProfitUnit instalProfitUnit = instalMeasure.getUnit(debtMoneyField);
                
                Map<String, AstProfitUnit> profitUnit = extractUnitForInstal(instalMeasure.getSubMeasure(),
                        debtMoneyField);
                profitSplitCalculator.splitProfitUnit(instalProfitUnit, profitUnit);
            }
            
            Map<String, LoanAstMeasure> instalSubMeasure = instalMeasure.getSubMeasure();
            // 3. 入账
            for (String loanAstNo : instalSubMeasure.keySet()) {
                LoanAstMeasure loanAstMeasure = instalSubMeasure.get(loanAstNo);
                loanAstMeasure.settle(instalTrans, memo);
            }
        }
    }

    private Map<Integer, InstalAstMeasure> buildInstalMeasure(LoanInstalDomain instalDomain,
                                                              LoanAstGroup astGroup, Date accountDate,Map<String,Money> loanAstPrinAmtMap) {
        Map<Integer, InstalAstMeasure> measureMap = Maps.newLinkedHashMap();
        // 根据机构资产转入日期排序
        Collections.sort(astGroup.getLoanAsts(), Comparator.comparing(LoanAst::getInDate));
        
        for (LoanAst loanAst : astGroup.getLoanAsts()) {
            
            LoanAstMeasure astMeasure = buildAstMeasure(instalDomain, loanAst, accountDate,loanAstPrinAmtMap.getOrDefault(loanAst.getLoanAstNo(),new Money()));
            
            Integer termNo = loanAst.getTermNo();
            
            InstalAstMeasure instalMeasure = measureMap.get(termNo);
            if (null == instalMeasure) {
                instalMeasure = InstalAstMeasure.createDummyByInstal();
                measureMap.put(termNo, instalMeasure);
            }
            
            instalMeasure.addSubMeasure(loanAst.getLoanAstNo(), astMeasure);
        }
        return measureMap;
    }

    private LoanAstMeasure buildAstMeasure(LoanInstalDomain srcInstal, LoanAst loanAst, Date accountDate,Money loanAstPrinAmt) {
        
        LoanAstProfit ownAstProfit = loanAst.getLoanAstProfit(srcInstal.getTermNo(), ProfitTypeEnum.OWN);
        if (null == ownAstProfit) {
            ownAstProfit = LoanAstUtil.buildDefaultProfit(srcInstal, loanAst, accountDate);
            
            // 定价为收益分润：需要在profit上设置定价
            if (loanAst.getAstPriceType() == PriceTypeEnum.SPLIT_PROFIT) {
                PlanBase plan = PlanContextHolder.getPlan(loanAst.getOwnOrg());
                
                if (plan == null) {
                    plan = planManager.getPlanFullByNo(loanAst.getOwnOrg());
                    PlanContextHolder.putPlan(plan);
                }
                
                PlanAstPriceRule priceRule = AstPriceUtil.filterPrice(plan.getPlanAstPriceRuleList(),
                        Long.valueOf(srcInstal.getTermNo()));
                
                if (null != priceRule) {
                    ownAstProfit.setSplitRate(priceRule.getIntrDiscountRate());
                }
                ownAstProfit.setSplitOrgCode(loanAst.getLenderOrgCode());
            }
        }

        LoanAstMeasure loanAstMeasure = LoanAstMeasure.buildByLoanAst(loanAst, srcInstal.getTermNo(),loanAstPrinAmt);
        return loanAstMeasure;
    }

    
    public  <T, P extends ProfitMeasure> Map<T, AstProfitUnit> extractUnitForInstal(Map<T, P> instalMeasure,
                                                              DebtMoneyFieldEnum fieldType) {
        Map<T, AstProfitUnit> ret = Maps.newLinkedHashMap();
        for (T key : instalMeasure.keySet()) {
            ProfitMeasure measure = instalMeasure.get(key);
            ret.put(key, measure.getUnit(fieldType));
        }
        return ret;
    }
}